#include "sa_types.h"
#include "sa_proto.h"

/* Take a combinatorial rectangle {K;v0,v1,v2,v3} and repack it
INSIDE a euclidean square, meaning that the boundary circles are
tangent to the square, the designated vertices in the corners.
Boundary vertices in counterclockwise order must be given in 
datastr. Return 0 on error.
*/

int alt_rect_pack(struct p_data *p,char *datastr)
{
  int i,v[4],v0,v1,v3,vert,click,nextv,success=0,passes=500,hits;
  int *oldnew=NULL,V[4],edgelength;
  char *endptr;
  struct Vertlist *vertlist=NULL,*trace;
  struct p_data *p1=NULL,*p2=NULL;

  /* check legality */
  if (!p->status || p->genus || (p->euler!=1 && p->euler!=2) ) {
    sprintf(msgbuf,"alt_rect_pack: improper specifications.");
    emsg();
    return 0; /* not simply connected */
  }
  if (!(vertlist=Node_link_parse(p,datastr,&endptr,&hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength)) || hits<4) {
    sprintf(msgbuf,"alt_rect_pack: didn't get four boundary vertices.");
    emsg();
    return 0;
  }
  trace=vertlist;
  for (i=0;(i<4 && trace);i++) {
    v[i]=trace->v;
    trace=trace->next;
  }
  vert_free(&vertlist);
  if (!p->packK_ptr[v[0]].bdry_flag || !p->packK_ptr[v[1]].bdry_flag 
      || !p->packK_ptr[v[2]].bdry_flag || !p->packK_ptr[v[3]].bdry_flag)
    goto BADVERTS;

  /* verts in counterclockwise order? */
  for (i=0;i<4;i++) {
    vert=v[i];
    click=-1;
    while ((nextv=p->packK_ptr[vert].flower[0])!=v[i]) {
      vert=nextv;
      click++;
      if ((i==0 && (vert==v[2] || vert==v[3]))
	  || (i==1 && vert==v[3])) {
      BADVERTS:
	sprintf(msgbuf,"alt_rect_pack: vertices improper or out of order");
	emsg();
	return 0;
      }
      else if (vert==v[(i+1)%4]) break; /* found next corner */
    } /* end of while */
    if (click<=0) {
      sprintf(msgbuf,"alt_rect_pack: one of identified sides is too short.");
      emsg();
      return 0;
    }
  }

  /* Procedure: intention is to build torus out of four copies of
     the quadrilateral with doubly reflective symmetry. This means
     there will be 2 doubling operations (but with ball bearings),
     then two adjoin operations to make the torus. */

  /* extra packings needed for temporary storage. */
  p1=(struct p_data *)NULL;
  p2=(struct p_data *)NULL;
  p1=init_packing();
  p2=init_packing();
  if (!p1 || !copy_pack(p,p1) || !copy_pack(p,p2)
      || !reverse_orient(p2)) {
    sprintf(msgbuf,"alt_rect_pack: failer to initiate tmp storsge.");
    emsg();
    goto BOMBOUT;
  }
  /* convert to euclidean */
  geom_to_e_pack_convert(p1);
  geom_to_e_pack_convert(p2);

  V[0]=v[0];V[1]=v[1];V[2]=v[2];V[3]=v[3];
  
  /* ==================== */
  /* Pasting 1: Add bearings along left end ({v0,v1} edge); then
     adjoin flipped copy. Add ball bearings along {v2,v3}; then
     adjoin to get cylinder. */

  if (!(edgelength=sq_grid_layer(p1,V,V+1))
      || !adjoin(p1,p2,V[1],v[1],edgelength,&oldnew)) {
    sprintf(msgbuf,"alt_rect_pack: some error in first pasting.");
    emsg();
    free(oldnew);
    goto BOMBOUT;
  }
  v3=oldnew[v[3]];
  free(oldnew);
  oldnew=NULL;


  if (!(edgelength=sq_grid_layer(p1,V+2,V+3))
      || !adjoin(p1,p1,V[3],v3,edgelength,&oldnew)) {
    sprintf(msgbuf,"alt_rect_pack: some error in second pasting.");
    emsg();
    free(oldnew);
    goto BOMBOUT;
  }
  v0=v[0]=oldnew[v[0]];
  v1=v[1]=oldnew[v[1]];
  free(oldnew);
  oldnew=NULL;

  /* Pasting 2: copy and reorient the whole packing. */

  if (!copy_pack(p1,p2) || !reverse_orient(p2)) {
    sprintf(msgbuf,"alt_rect_pack: copy or reorient area.");
    emsg();
    goto BOMBOUT;
  }
  if (!(edgelength=sq_grid_layer(p1,v,v))
    || !adjoin(p1,p2,v[0],v0,edgelength,&oldnew)) {
    sprintf(msgbuf,"alt_rect_pack: some error in third pasting.");
    emsg();
    free(oldnew);
    goto BOMBOUT;
  }
  v1=oldnew[v1];
  free(oldnew);
  oldnew=NULL;
  if (!(edgelength=sq_grid_layer(p1,v+1,v+1))
      || !adjoin(p1,p1,v[1],v1,edgelength,&oldnew)) {
    sprintf(msgbuf,"alt_rect_pack: some error in fourth pasting.");
    emsg();
    free(oldnew);
    goto BOMBOUT;
  }
  free(oldnew);
  oldnew=NULL;

  /* should now have the new packing in p1. */
  passes=500;
  if (p1->nodecount>500) passes=2*(p1->nodecount);
  fix_call(p1,"-F");
  int erif;
  erif=e_riffle(p1,passes);

  /* copy radii to p */
  geom_to_e_pack_convert(p);
  for (i=1;i<=p->nodecount;i++) p->packR_ptr[i].rad=p1->packR_ptr[i].rad;
  comp_pack_centers(p,0,0,2,okerr);
  success=1;

 BOMBOUT:
  free_p_data(&p2);
  free_p_data(&p1);
  return success;
} /* alt_rect_pack */





