#include "sa_types.h"
#include "sa_proto.h"
#include "sa_head.h"

/* Routine for setting ratio function for boundary vertices
using a prescribed function and various options and parameters. 
Idea is to set up iterative process to build discrete versions of 
Beurling's extension of Riemann's mapping thm by prescribing the 
ratio function for bdry verts. The ratio function values are
obtained with calls to CirclePack, where the "Function" tab 
function is used. The calling routine must do repack, etc. 

  Beurling's generalization of Riemann mapping theorem says that
given bounded, continuous positive function h on the plane and log(1/h)
is subharmonic, then there exists a univalent analytic function f so 
that |f'(z)| - h(f(z)) --> 0 as z --> e^{it} on unit circle and
f is unique if f(0)=0 and f'(0)>0. 

pr = range pack (this one gets modified) and pd = domain pack.
pr must be eucl (for now); if pd is hyp, radii will be converted 
to eucl for use with ratio functions.

t: This parameter t is normally in [0,1] and moderates the
change. In particular, if desired ratio function is d and
current ratio function is c, then set ratio function to
exp(t*log(d/c))*c. Thus, t=0 means no change, t=1 means replace
c by d, and for 0<t<1, interpolate.

xpnt: This is a positive exponent for function h, default xpnt=1.0;
replaces d by d^x. (This somehow might mimic the exponent on phi'(z) 
in conformal flow paper (see Collins, Driscoll, Stephenson).

w0: translation (accomplished by shifting argument of h) by complex
number w0; default is 0.

option: May add options, but for now:
   0: (default) adjust all boundary
   1: only adjust if d (desired) is greater than c (current).
   2: only adjust if d is less than c.

Return count of radii set, 0 on error.
*/


int beurling_flow(struct p_data *pr,struct p_data *pd,
		  double t,double xpnt,complex w0,int option)
{
  int v,count=0,transflag=0,node;
  double CPhz,e_rad,cur,factor=1.0,*ratio_data=NULL,rat;
  complex z,e_ctr;
  char outfile[NAME_MAX];
  struct K_data *pKr=pr->packK_ptr;
  struct R_data *pRd=pd->packR_ptr,*pRr=pr->packR_ptr;
  FILE *rfp=NULL;

  if (!pd->status || !pr->status || pr->hes || pd->hes>0
      || pd->nodecount < pr->nodecount || t<0 ) {
    sprintf(msgbuf,"usage: beurling_flow: check conditions.");
    emsg();
sprintf(msgbuf,"pd->status %d, pr->status %d, pr->hes %d, pd->hes %d, pd->nodecount %d, pr->nodecount %d, t=%lf",
pd->status,pr->status,pr->hes,pd->hes,pd->nodecount,pr->nodecount,t);
debugmsg();
    return 0;
  }
  if (w0.re!=0.0 || w0.im!=0.0) transflag=1;

  /* cycle through bdry pr */
  for (v=1;v<=pr->nodecount;v++)
    if (pKr[v].bdry_flag) {
      z=pRr[v].center;
      if (transflag) z=cadd(z,w0);
      if (pd->hes<0) /* hyp: convert rad */
	h_to_e_data(pRd[v].center,pRd[v].rad,&e_ctr,&e_rad);
      else e_rad=pRd[v].rad;
	
	CPhz=cAbs(get_java_ftn_value(z));
      
	/* Set the new radius */
     
    if (xpnt!=1.0) CPhz=exp(xpnt*log(CPhz));
    cur=pRr[v].rad/e_rad;
    factor=exp(t*log(CPhz/cur));
    if (option==1) {
	if (factor>1.0) /* increases only */
	  pRr[v].rad *= factor;
      }
    else if (option==2) {
	if (factor<1.0) /* decreases only */
	  pRr[v].rad *= factor;
      }
    else
	pRr[v].rad *= factor;
      count++;
    }
  if (ratio_data) free(ratio_data);
  return count;
} /* beurling_flow */
