#include "sa_types.h"
#include "sa_proto.h"

/* Code to convert strings of H's and V's into Brooks packings.
This involves modification of a quadrilateral interstice (i.e.,
centered on a 4-degree vertex) by successive addition of 
'horizontal' and 'vertical' circles a la Robert Brooks, Duke Math.
J. 1985. Final circle is a 'plug'. (Generically, Brooks packings
go on ad infinitum.). Original routine conceived by Matt Cathey.

There are two modes of operation indicated by flag: 
  (1) insert new circle in location as in Brooks' original scheme; 
  (2) insert new circle in alternating fashion: verticals alternate
      left/right, horizontals, up/down. */

#define MAX_LIST 4000   /* limit on list of H's and V's */

int brooks_construct(struct p_data *p,int v,int w,
		    struct Vertlist *VH_list,int flag,int cflag)
     /* v must be interior, 4-degree vertex, w is neighbor (thought
of as 'top' in the Brooks strategy), VH_list is linked list of 1's (V's)
and 2's (H's), flag: 1 = original Brooks protocol, 2 = "alternating" 
protocol. If cflag set, color V circles red, H circles blue. */
{
  int T,R,B,L,dum,N,P;
  int v_flips=1,h_flips=0,h_mode=0,v_mode=0;
  int i,j,k;
  char *c;
  struct Vertlist *vtrace;
  struct K_data *pK_ptr=p->packK_ptr;

  if (pK_ptr[v].bdry_flag || pK_ptr[v].num!=4 
      || (j=nghb(p,v,w))<0 || !VH_list) return 0;

  /* initialize the variables we need to track: T,R,B,L are the
     circles forming the current quadrilateral. When in flag=2,
     we flip (interchange T/B, L/R) each time we add a new vertex. 
     N is the number of the old plug (now to become the new
     vertex), P is incremented for the new plug. */

  T=pK_ptr[v].flower[j];
  L=pK_ptr[v].flower[(j+1)%(pK_ptr[v].num)];
  B=pK_ptr[v].flower[(j+2)%(pK_ptr[v].num)];
  R=pK_ptr[v].flower[(j+3)%(pK_ptr[v].num)];
  N=v;
  P=p->nodecount; 

  if (flag==2) /* prepare for first flip */
    {dum=T;T=B;B=dum;dum=L;L=R;R=dum;}
   
  /* Cycle through reading V's and H's (1's and 2's) from list;
     stop on reaching MAX_LIST cap. */

  k=0;
  vtrace=VH_list;
  while (k<MAX_LIST && vtrace)
    {
      P++; 
      if (P>p->sizelimit && !alloc_pack_space(p,P,1))
	return 0; /* failed to allocate new pack space */
      pK_ptr=p->packK_ptr;
      p->nodecount=P;
      if (vtrace->v==1) /* next circle is 'vertical' */
	{
	  if (!v_mode) v_flips=0;
	  if (flag==2 && (h_flips%2)==0)
		/* Flip (exchange labels) each V. However, for first 
		   V in chain, may have to compensate for flips done
		   while in previous horizontal chain. */
	    {
	      dum=T;T=B;B=dum; /* exchange T and B */
	      dum=L;L=R;R=dum; /* exchange L and R */
	      v_flips++;
	    }
	  h_mode=0;
	  v_mode=1;
	  brooks_insert(p,P,N,T,-1);
	  brooks_insert(p,P,N,B,1);
	  
	  /* fix flower of R */
	  for (i=0;i<=pK_ptr[R].num;i++)
	    if (pK_ptr[R].flower[i]==N) pK_ptr[R].flower[i]=P;
	  
	  /* fix flower of N */
	  pK_ptr[N].flower[0]=T;
	  pK_ptr[N].flower[1]=L;
	  pK_ptr[N].flower[2]=B;
	  pK_ptr[N].flower[3]=P;
	  pK_ptr[N].flower[4]=T;
	  if (cflag) pK_ptr[N].color = 150; /* red */
	  
	  /* add flower of new temporary 'plug' vert */
	  pK_ptr[P].flower=(int *)calloc(5,sizeof(int));
	  pK_ptr[P].num=4;
	  pK_ptr[P].bdry_flag=0;
	  p->packR_ptr[P].rad=.5;
	  p->packR_ptr[P].rad=2.0*M_PI;	  
	  pK_ptr[P].flower[0]=N;
	  pK_ptr[P].flower[1]=B;
	  pK_ptr[P].flower[2]=R;
	  pK_ptr[P].flower[3]=T;
	  pK_ptr[P].flower[4]=N;
	  
	  L=N; /* N is new 'left' of interstice */
	  
	}
      else if (vtrace->v==2) /* next circle is 'horizontal' */
	{
	  if (!h_mode) h_flips=0;
	  if (flag==2 && (v_flips%2)==0)
		/* Flip (exchange labels) each H. However, for first 
		   H in chain, may have to compensate for flips done
		   while in  chain. */
	    {
	      dum=T;T=B;B=dum; /* exchange T and B */
	      dum=L;L=R;R=dum; /* exchange L and R */
	      h_flips++;
	    }
	  v_mode=0;
	  h_mode=1;
	  brooks_insert(p,P,N,R,-1);
	  brooks_insert(p,P,N,L,1);
	  
	  /* fix flower of 3 */
	  for (i=0;i<=pK_ptr[B].num;i++)
	    if (pK_ptr[B].flower[i]==N) pK_ptr[B].flower[i]=P;
	  
	  /* fix flower of N */
	  pK_ptr[N].flower[0]=T;
	  pK_ptr[N].flower[1]=L;
	  pK_ptr[N].flower[2]=P;
	  pK_ptr[N].flower[3]=R;
	  pK_ptr[N].flower[4]=T;
	  if (cflag) pK_ptr[N].color = 50; /* blue */

	  /* add flower of new P vert */
	  pK_ptr[P].flower=(int *)calloc(5,sizeof(int));
	  pK_ptr[P].num=4;
	  pK_ptr[P].bdry_flag=0;
	  p->packR_ptr[P].rad=.5;
	  p->packR_ptr[P].rad=2.0*M_PI;	  
	  pK_ptr[P].flower[0]=N;
	  pK_ptr[P].flower[1]=L;
	  pK_ptr[P].flower[2]=B;
	  pK_ptr[P].flower[3]=R;
	  pK_ptr[P].flower[4]=N;
	  
	  T=N; /* N is new 'top' of interstice */
	}
      
      k++;
      N=P;
      vtrace=vtrace->next;
    } /* end of while */
  complex_count(p,0);
  facedraworder(p,0);
  set_aim_default(p);
  return k;
} /* brooks_construct */

int brooks_insert(struct p_data *p,int P,int N,int v,int flag)
     /* insert new P in flower v before (flag=1) or after (flag=-1) N */
{
  int i,j,done=0,*newflower,num;
  struct K_data *pK_ptr=p->packK_ptr;

  num=pK_ptr[v].num;
  newflower=(int *)calloc(num+2,sizeof(int));
  for (i=0;i<=num;i++)
    newflower[i]=pK_ptr[v].flower[i];
  if (flag==1) /* insert P before N */
    {
      for (i=0;(i<=num) && !done;i++)
	{
	  if (pK_ptr[v].flower[i]==N)
	    {
	      for (j=i;j<=num;j++)
		newflower[j+1]=pK_ptr[v].flower[j];
	      newflower[i]=P;
	      done=1;
	    }
	}
      if (!done) return 0; /* error: should have found M */
    }
  else /* insert P after N */
    {
      for (i=0;(i<=num) && !done;i++)
	{
	  if (pK_ptr[v].flower[i]==N)
	    {
	      for (j=i+1;j<=num;j++)
		newflower[j+1]=pK_ptr[v].flower[j];
	      newflower[i+1]=P;
	      done=1;
	    }
	}
      if (!done) return 0; /* error: should have found M */
    }
  pK_ptr[v].num++;
  free(pK_ptr[v].flower);
  pK_ptr[v].flower=newflower;
  return 1;
} /* brooks_insert */







