#include "sa_types.h"
#include "sa_proto.h"
#include "Interface.h"

/* Find circle centers for a packing based on current radii by 
going through faces in precomputed order (see facedraworder).
However, have various options and error cutoffs to improve quality
and catch errors. (Note that spherical layout is not very robust;
generally best to project from plane or disc.)

  There is a 'plot_flag' for each vertex; it is set if placement 
seems okay; some display/print situations may use only circles 
with plot_flag>0.

  Start layout with 'first_face' (normally, 'alpha' at origin); last
step normalizes 'gamma' vert on y>0 axis. Only compute each 
circle's center once. If errflag, only circles placed w/o too much 
error have plot_flag>0. 

  Wrap up tasks:
  * identify/fix any 'nan' centers, which happen most often in
  inv dist cases. For now, fix by using average of nghb centers.

  * Record rad/cent in redlist, if it exists, BUT use cent/rad 
  data from pack (not redlist data).

  * Use redlist data to set any side-pairing Mobius transforms.

Options: 
 errflag true: plot_flag <=0 if center doesn't meet placement
             criteria; use only verts with plot_flag's >0 
	     when computing later centers.

 dflag true: write diagnostic file /tmp/layout.log of problem placements.

 opt: 1 = use drawing-order only (ie. go to next face, compute
          center of third vert from other two -- this is the 
	  original 'fix' method).
      2 = use average of already plotted neighbors
      3 = ???

 crit: this double gives accuracy criteria (if errflag set). 
       Eg., don't use computed center of v if nghbs u,w actual 
       distance apart is off by factor > crit. Should have default 
       value to use for crit in general; can experiment with it.

Return number of failed placements.

*/

int comp_pack_centers(struct p_data *p,int errflag,int dflag,int opt,
		     double crit)
{
  int nf,i,j,n0,n1,n,vert,count,v,indx,lastface,flick=1;
  int v_ind,wflag=0,debug=0,simp_conn_flag=0;
  complex z;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;
  f_data *faces=p->faces;
  struct RedList *trace;
  FILE *diagfp=NULL;

  if ((p->euler==1 || p->euler==2) && p->genus==0) simp_conn_flag=1;
  for (j=1;j<=p->nodecount;j++) pK_ptr[j].plot_flag=pK_ptr[j].qual_flag=0;
  for (j=1;j<=p->facecount;j++) faces[j].plot_flag=1;
  nf=p->first_face;

  if (dflag && errflag) {
    if (!(diagfp=fopen("/tmp/layout.log","a"))) {
      sprintf(msgbuf,"couldn't open /tmp/layout.log");
      emsg();
      return 0;
    }
    fprintf(diagfp,"Diagnostic file created in 'new_pack_centers'.\n"
	    "Packing %s, nodecount %d, 'critical' tolerance set at %f\n\n"
	    "Following are failed placements:\n"
	    "\tat count \tvert \tface\n\n",
	    p->file_name,p->nodecount,(singledouble)crit);
  }

  /* bomb out if first face fails to lay out properly */

  if (!place_face(p,nf,faces[nf].index_flag)) {
    sprintf(msgbuf,"Layout error in placing the initial face.");
    emsg();
    if (diagfp) {
      fprintf(diagfp,"first face, %d problem.\n",nf);
      fclose(diagfp);
      return 0;
    }
    return 0;
  }
   
  for (j=0;j<3;j++) pK_ptr[faces[nf].vert[j]].plot_flag=1;
  count=3;

/* ------------ simply connected case ----------- */

  if (simp_conn_flag) {
    while ( (nf=faces[nf].next_face)!=p->first_face 
	    && nf>0 && nf<=p->facecount) {
      vert=faces[nf].vert[(indx=(faces[nf].index_flag +2) % 3)];
      n0=nghb(p,vert,faces[nf].vert[(indx+1)%3]);
      if (pK_ptr[vert].plot_flag<=0 && (pK_ptr[vert].plot_flag=
	  fancy_comp_center(p,vert,n0,0,pK_ptr[vert].num,
			    errflag,opt,0,crit)))
	count++;
      else if (pK_ptr[vert].plot_flag<=0 && dflag && errflag)
	fprintf(diagfp,"\t%d \t%d \t%d   \n",count,vert,nf);
    }
  } /* end of first pass in simply connected case  */

/* -------- multiply connected case ----------------- */

  else { 
    for (i=1;i<=p->nodecount;i++) pK_ptr[i].util_flag=0;

    if (opt==2) { /* want to use average, so have to mark outer verts 
		     of red chain using util_flag; need to find fans. */
      trace=p->redfaces;
      wflag=0;
      while (trace!=p->redfaces || !(wflag++)) {
	indx=nghb_tri(p,trace->next->face,trace->face);
	v=p->faces[trace->face].vert[indx];
	if (!pK_ptr[v].util_flag) pK_ptr[v].util_flag=1;
	if (trace->next->face==trace->prev->face) { /* blue? */
	  indx=nghb_tri(p,trace->next->face,trace->face);
	  v=p->faces[trace->face].vert[(indx+2) % 3];
	  pK_ptr[v].util_flag=1;
	}
	trace=trace->next;
      }
    }

    lastface=p->first_face;
    nf=p->first_face;
    if (p->faces[p->first_face].next_face==p->first_red_face)
      nf=p->first_red_face;
    while (nf>0 && nf<=p->facecount && count<3*p->facecount) {
      if (flick && nf==p->first_red_face) {
	/* When we first reach first_red_face, we switch to layout 
	   whole red chain in order. This is:
	   1) necessary because some faces will occur twice, placing 
	   possibly a different one of its circles each time, and 
	   2) to load rad/cent data in redfaces data structure
	   for other uses. 
	   Note: circles end up wherever last placed. */
	
	flick=0;
	vert=faces[nf].vert[(indx=(faces[nf].index_flag +2) % 3)];
	n0=nghb(p,vert,faces[nf].vert[(indx+1)%3]);
	if (pK_ptr[vert].util_flag) /* this is redchain outer vert
				       and we're using averaging. */
	  red_fan(p,p->redfaces,vert,&n1,&n);
	else {
	  n1=0;
	  n=pK_ptr[vert].num;
	}
	if (pK_ptr[vert].plot_flag<=0 && (pK_ptr[vert].plot_flag=
	    fancy_comp_center(p,vert,n0,n1,n,errflag,opt,0,crit)))
	  count++;
	else if (dflag && errflag && pK_ptr[vert].plot_flag<=0)
	  fprintf(diagfp,"\t%d \t%d \t%d   \n",count,vert,nf);

	p->redfaces->center=
	  p->packR_ptr[p->faces[p->redfaces->face].
		       vert[p->redfaces->v_flag]].center;
	p->redfaces->rad=
	  p->packR_ptr[p->faces[p->redfaces->face].
		       vert[p->redfaces->v_flag]].rad;

	/* Now proceed through rest of redfaces.
	   Which center to compute is based on indices stored 
	   in f_data. This should fit 'next_face' order (which 
	   reflects redfaces order). Note, this can differ 
	   from v_flag in redlist. Record rad/cent whether 
	   or not error is within tolerance. */
	
	/* This is very confusing. Only problem with going
	   through redfaces in order, each depending only on
	   last, is that errors will build up and bdry may
	   detach from white faces. Indices in f_data try to
	   be smart in choosing which of three verts to plot,
	   but I'm not sure if it might rely on outdated 
	   locations for the other two. May have to pad the
	   'next_face' order with sections of redfaces list
	   which it skips (because those faces already had
	   a 'next_face' -- see 'final_order'). */
	
	lastface=p->first_red_face;
	trace=p->redfaces->next;
	while (trace!=p->redfaces) {
	  nf=trace->face;
	  if ((indx=nghb_tri(p,lastface,nf))<0) {
	    sprintf(msgbuf,"Problem with redfaces order.");
	    emsg();
	    lastface=nf;
	    goto CARRYON;
	  }
	  indx=(p->faces[nf].index_flag+2) % 3;
	  vert=p->faces[nf].vert[indx];
	  n0=nghb(p,vert,faces[nf].vert[(indx+1)%3]);
	  if (pK_ptr[vert].util_flag) /* this is redchain outer vert
					 and we're using averaging. */
	    red_fan(p,trace,vert,&n1,&n);
	  else {
	    n1=0;
	    n=pK_ptr[vert].num;
	  }
	  if ((pK_ptr[vert].plot_flag=
	       fancy_comp_center(p,vert,n0,n1,n,0,1,0,crit)))
	    /* Note: since the red faces are needed downstream
	       and it's difficult to keep track of which 
	       nghbs should be used to plot a vert at each visit,
	       we use values errflag=0 and opt=1 in this call. */
	    
	    /* fixup: Can we rearrange drawing order so as many
	       verts as possible not on outside of red chain 
	       are laid out first, then the red chain? */
	    
	    count++;
	  else if (dflag && errflag)
	    fprintf(diagfp,"\t%d \t%d \t%d   \n",count,vert,nf);
	  
	  /* Now record rad/cent in redface data. This may not 
	     be the data we just computed, so we may have to
	     do yet another computation here (depends on v_flag). */
	  
	  v_ind=p->faces[trace->face].vert[trace->v_flag];
	  if (v_ind!=vert) { /* need to update center for circle
				associated with v_flag of redface. */
	    n0=nghb(p,v_ind,faces[trace->face].
		    vert[((trace->v_flag)+1)%3]);
	    fancy_comp_center(p,v_ind,n0,0,0,0,1,1,crit);
	  }
	  trace->center=p->packR_ptr[v_ind].center;
	  trace->rad=p->packR_ptr[v_ind].rad;
	  
	  lastface=nf;
	  trace=trace->next;
	}
	/* debug */
	if (debug) red_bug(p,p->redfaces);
      }

      /* done with red chain; pick up order from f_data again */

    CARRYON: 
      if ((nf=p->faces[lastface].next_face)==p->first_face) 
	break;
      vert=faces[nf].vert[(indx=(faces[nf].index_flag +2) % 3)];
      n0=nghb(p,vert,faces[nf].vert[(indx+1)%3]);
      if (pK_ptr[vert].plot_flag<=0 && (pK_ptr[vert].plot_flag=
	  fancy_comp_center(p,vert,n0,0,pK_ptr[vert].num,
				 errflag,opt,0,crit)))
	count++;
      else if (dflag && errflag && pK_ptr[vert].plot_flag<=0)
	fprintf(diagfp,"\t%d \t%d \t%d   \n",count,vert,nf);

      lastface=nf;
      nf=p->faces[lastface].next_face;
    }
  } /* end of multiply connected case  */

  if (dflag && errflag)
    fprintf(diagfp,"\nFinished original pass with %d successful"
	    " placements, nodecount %d.\n\n",count,p->nodecount);

/* Check whether some circles weren't placed. Make several passes to 
   catch any further legitimate layouts. Note: count may exceed nodecount
   in multiply-connected cases since a center may be computed several times */

  wflag=0;
  for (i=1;i<=p->nodecount;i++)
    if (!pK_ptr[i].plot_flag) wflag++;
  if (wflag) { /* Some circles not plotted */
    wflag=1;
    while (wflag) { /* with hits, keep cycling through */
      wflag=0;
      nf=p->first_face;
      while ((nf=faces[nf].next_face)!=p->first_face) {
	vert=faces[nf].vert[(indx=(faces[nf].index_flag +2+j) % 3)];
	if (pK_ptr[vert].plot_flag<=0) {
	  n0=nghb(p,vert,faces[nf].vert[(indx+1)%3]);
	  if ((pK_ptr[vert].plot_flag=
	       fancy_comp_center(p,vert,n0,0,pK_ptr[vert].num,
				 errflag,opt,0,crit))) {
	    count++;
	    wflag++;
	  }
	}
      } /* end of inner while */
    } /* end of outer while */

    /* Try to find appropriate center
       for circles that didn't plot for some reason (eg. inv dist
       incompatibility). Simply average the centers of any petals
       that are plotted (often problem circles are extremely small,
       so this is reasonable). I'll use 'qual_flag' to keep track,
       and pass through repeatedly as long as we're getting results. 
       (Some centers may be based on neighbors which were only set 
       based on their neighbors, but at least they're getting some
       location!) I do not set their plot_flags. */

    wflag=1;
    while (wflag) {
      wflag=0;
      for (vert=1;vert<=p->nodecount;vert++) {
	if (!pK_ptr[vert].plot_flag && !pK_ptr[vert].qual_flag) {
	  z.re=z.im=0.0;
	  n=0;
	  /* first use plotted neighbors */
	  for (j=0;j<=pK_ptr[vert].num+pK_ptr[vert].bdry_flag;j++) 
	    if (pK_ptr[(v=pK_ptr[vert].flower[j])].plot_flag) {
	      z=cadd(z,pR_ptr[v].center);
	      n++;
	    }
	  if (!n) { /* none? try qual_flag neighbors */
	    for (j=0;j<=pK_ptr[vert].num+pK_ptr[vert].bdry_flag;j++) 
	      if (pK_ptr[(v=pK_ptr[vert].flower[j])].qual_flag) {
		z=cadd(z,pR_ptr[v].center);
		n++;
	      }
	  }
	  if (n) { /* got something! */
	    z.re /= n;
	    z.im /= n;
	    pR_ptr[vert].center=z;
	    pK_ptr[vert].qual_flag=1;
	    count++;
	    wflag++;
	  }
	}
      } /* end of for */
    } /* end of while */
  }

  if (simp_conn_flag && (trace=p->redfaces)) {
    wflag=0;
    while (trace!=p->redfaces || !(wflag++)) {
      v=p->faces[trace->face].vert[trace->v_flag];
      trace->center=p->packR_ptr[v].center;
      trace->rad=p->packR_ptr[v].rad;
      trace=trace->next;
    }
  }

  /* Set face plot_flags to default */
  for (j=1;j<=p->facecount;j++) faces[j].plot_flag=1;

  /* For simply connected, try this; precision is rather experimental. 
   fixup: this is much more complicated in multiply conn case. */

  if (simp_conn_flag && (trace=p->redfaces) && errflag) {
    for (j=1;j<=p->facecount;j++)
      if (pK_ptr[faces[j].vert[0]].plot_flag<=0
	  || pK_ptr[faces[j].vert[1]].plot_flag<=0
	  || pK_ptr[faces[j].vert[2]].plot_flag<=0
	  || fabs(invdist_err(p,faces[j].vert[0],faces[j].vert[1]))
	  >100000*toler
	  || fabs(invdist_err(p,faces[j].vert[1],faces[j].vert[2]))
	  >100000*toler
	  || fabs(invdist_err(p,faces[j].vert[2],faces[j].vert[0]))
	  >100000*toler)
	faces[j].plot_flag=0;
  }
    
/* Normalize: fixup? what if errflag is set and gamma is poorly plotted? */

  norm_any_pack(p,p->packR_ptr[p->alpha].center,
		p->packR_ptr[p->gamma].center); 
  
/* Record any edge-pairing Mobius transforms */
  update_pair_mob(p);
//sprintf(msgbuf,"run updateSidePairs");
//debugmsg();
  updateSidePairs(p); // send to Java

  if (diagfp) fclose(diagfp);
  return (count);

} /* comp_pack_centers */

