#include "sa_types.h"
#include "sa_proto.h"

/* A circle C=C(r,z) whose disc is in the puncture plane has 
   aspect ratio AR(C)= r/|z|. Any triple of eucl circles generates 
   a doyle spiral winds around some point p in the plane. 
   This computes the point p and the doyle parameters 
   a, b (see "spiral") associated with its circles. Return 0
   on error. Note, we do not check that the given circles are in 
   fact tangent.*/

int doyle_point(double r0,double r1,double r2,
		complex z0,complex z1,complex z2,
		double *A,double *B,complex *p)
{
  int j,J=0,n;
  double a,b,least=0.0,rad[3],theta,N,coef,holdr;
  complex cent[3],accum,w,term,holdz;

  rad[0]=r0;cent[0]=z0;rad[1]=r1;cent[1]=z1;rad[2]=r2;cent[2]=z2;
  least=r0;
  for (j=1;j<3;j++) 
    if (rad[j]<least) {least=rad[j];J=j;}
  if (least<=.00000000001) return 0; /* too small to work with */
  /* shift so 0 index has smallest rad */
  if (J==1) {
    holdz=cent[1];cent[1]=cent[2];cent[2]=cent[0];cent[0]=holdz;
    holdr=rad[1];rad[1]=rad[2];rad[2]=rad[0];rad[0]=holdr;
  }
  else if (J==2) {
    holdz=cent[2];cent[2]=cent[1];cent[1]=cent[0];cent[0]=holdz;
    holdr=rad[2];rad[2]=rad[1];rad[1]=rad[0];rad[0]=holdr;
  }
  *A=a=rad[1]/rad[0];
  *B=b=rad[2]/rad[0];

  /* Computation via truncated infinite series is thanks to Matt Cathey. */
  theta=acos((1 + a + b - a*b) / (1 + a + b + a*b))
    + acos((a + a*b + b - b*b) / (a + a*b + b + b*b))
    + acos((a*a + a + a*b - b) / (a*a + a + a*b + b));
  accum.re=accum.im=0.0;
  for (n=1;n<=100;n++) {
    N=(double)n;
    coef=pow(a,(1-N))+pow(a,-N);
    w=cExp(1.0,N*theta-(N-1)*M_PI);
    term.re=coef*w.re;
    term.im=coef*w.im;
    accum = cadd(accum,term);
  }
  /* scale back to original size, translate to reestablish center cent[0]. */
  accum.re *= rad[0];
  accum.im *= rad[0];
  *p=cadd(accum,cent[0]);
  return 1;
} /* doyle_point */

int doyle_point_call(struct p_data *p,char *datastr)
{
  int f;
  double a,b,r0,r1,r2;
  complex z,z0,z1,z2;

  if (!p->status || p->hes
      || !(f=grab_one_face(p,&datastr))) {
    sprintf(msgbuf,"doyle_point: must specify face f of eucl packing.");
    emsg();
    return 0;
  }
  r0=p->packR_ptr[p->faces[f].vert[0]].rad;
  r1=p->packR_ptr[p->faces[f].vert[1]].rad;
  r2=p->packR_ptr[p->faces[f].vert[2]].rad;
  z0=p->packR_ptr[p->faces[f].vert[0]].center;
  z1=p->packR_ptr[p->faces[f].vert[1]].center;
  z2=p->packR_ptr[p->faces[f].vert[2]].center;
  if (!doyle_point(r0,r1,r2,z0,z1,z2,&a,&b,&z)) return 0;
  sprintf(msgbuf,"doyle_point for face %d: p = %lf + i * %lf, "
	  "parameters: a = %lf, b = %lf.",f,
	  (singledouble)z.re,(singledouble)z.im,
	  (singledouble)a,(singledouble)b);
  msg();
  return 1;
} /* doyle_point_call */
    

    
    
    
