/* eucl_math.c    euclidean geometry utility routines 
Ken Stephenson, 8/91 */

/* cc eucl_math.c -o eucl_math.o -c -g -lm */

#include "sa_types.h"
#include "sa_proto.h"
extern int s_pt_to_vec();

double e_area(double r1,double r2,double r3)
     /* return area of triangle given radii */
{
  double s;
	
  s=(r1+r2+r3);
  return (sqrt(s*(s-r1-r2)*(s-r1-r3)*(s-r2-r3)));
} /* e_area */

complex eucl_tri_center(complex p1,complex p2,complex p3)
{
  complex ctr;

  ctr.re=(p1.re+p2.re+p3.re)*.3333333;
  ctr.im=(p1.im+p2.im+p3.im)*.3333333;
  return ctr;
} /* eucl_tri_center */
 
int e_compcenter(complex z1,complex z2,complex *z3,
		 double e1,double e2,double *e3,
		 double o1,double o2,double o3)
     /* given centers/radii of circles 1 and 2 and rad of 3, and 
inv distances oj opposite zj, find eucl center of 3. (e3 is a pointer 
only for compatibility reasons.) */ 
{
  double l1,l2,l3,ld,sq23,s;
  complex v,w;
  extern double cAbs(complex); 
	
  l3=e1*e1+e2*e2+2*e1*e2*o3;
  l2=e1*e1+(*e3)*(*e3)+2*e1*(*e3)*o2;
  l1=e2*e2+(*e3)*(*e3)+2*e2*(*e3)*o1;
  ld=0.5*(l2+l3-l1);
  if ((ld*ld)>(l2*l3))
    return 0;
  v.re=ld/sqrt(l3); /* ld/(2*sqrt(l2*l3)) = cos(angle) */
  v.im=sqrt(l2-v.re*v.re);
  w.re=z2.re-z1.re;
  w.im=z2.im-z1.im;
  s = cAbs(w);
  if (s<1.0e-30) return 0;
  w.re = w.re/s; w.im = w.im/s;
  z3->re=z1.re+w.re*v.re-w.im*v.im;
  z3->im=z1.im+w.re*v.im+w.im*v.re;
  return 1;
} /* e_compcenter */
	
double e_cos_overlap(double e1,double e2,double e3,
		     double t1,double t2,double t3,int *flag)
     /* given three eucl radii, cos's of opp overlap angles, return 
cos of angle at e1. If radii and overlaps incompatible, return 
value (for 0 or Pi, depending) and set flag. Note: tj <= 1 for 
overlap, tj > 1 for inv distance.  */
{
        double l2,l3,e12,e22,e32,x1,x2,x3,sqrlen;

	e12 = e1*e1;
	e22 = e2*e2;
	e32 = e3*e3;
	x3 = e1*e2*t3;
	x2 = e1*e3*t2;
	x1 = e2*e3*t1;
	l3=e12+e22+2*x3;
	l2=e12+e32+2*x2;
	sqrlen=(e12+x2+x3-x1)/sqrt(l2*l3);
	if (sqrlen < -1.0){(*flag)++; return -1.0;} /* error? */
	if (sqrlen > 1.0){(*flag)++; return 1.0;} /* error? */
	return sqrlen;

} /* e_cos_overlap */

double R3_dist(complex z,complex w)
     /* distance in R^3 between sph. pts z and w */
{
  double Z[3],W[3];
  
  s_pt_to_vec(z,Z);
  s_pt_to_vec(w,W);
  
  Z[0]=Z[0]-W[0];
  Z[1]=Z[1]-W[1];
  Z[2]=Z[2]-W[2];
  
  return (vec_norm(Z));
} /* R3_dist */

double xyz_dist(struct xyzPoint xyz,struct xyzPoint XYZ)
     /* distance in three space from (x,y,z) to (X,Y,Z) */
{
  return (sqrt((xyz.x-XYZ.x)*(xyz.x-XYZ.x)+
	       (xyz.y-XYZ.y)*(xyz.y-XYZ.y)+
	       (xyz.z-XYZ.z)*(xyz.z-XYZ.z)));
} /* xyz_dist */
  

double e_cos_overlap_special(double e1,double e2,double e3,
			     double t1,double t2,double t3,
			     int *flag,double *incompat_err)
     /* given three eucl radii, cos's of opp overlap angles, return 
cos of angle at e1. If radii and overlaps incompatible, return 
value (for 0 or Pi, depending) and set flag. Note: tj <= 1 for 
overlap, tj > 1 for inv distance. */
{
	double l2,l3,e12,e22,e32,x1,x2,x3,sqrlen;

	e12 = e1*e1;
	e22 = e2*e2;
	e32 = e3*e3;
	x3 = e1*e2*t3;
	x2 = e1*e3*t2;
	x1 = e2*e3*t1;
	l3=e12+e22+2*x3;
	l2=e12+e32+2*x2;
	sqrlen=(e12+x2+x3-x1)/sqrt(l2*l3);

  if (sqrlen < -1.0)
    {
      (*flag)++;
      *incompat_err += fabs(sqrlen);
      return -1.0;
    } /* error? */
  if (sqrlen > 1.0)
    {
      (*flag)++;
      *incompat_err += sqrlen;
      return 1.0;
    } /* error? */ 
  return sqrlen;
} /* e_cos_overlap_special */

double Fx(double x,double y,double z)
     /* deriv in eucl setting for angle at x w.r.t radius x.*/
{
  double a,b;

  a=x*(x+y+z); b=y*z;
  return ( (-b)*(2*x+y+z)/((a+b)*sqrt(a*b)) );
} /* Fx */
	
double Fy(double x,double y,double z)
     /* deriv in eucl setting for angle at x w.r.t radius y.*/
{
  double a,b;

  a=x*(x+y+z); b=y*z;
  return (x*y*(x+y)/((a+b)*sqrt(a*b)));
} /* Fy */

double Fz(double x,double y,double z)
     /* deriv in eucl setting for angle at x w.r.t radius y.*/
{
  return (Fy(x,z,y));
} /* Fz */	 
	
double inv_dist(complex z1,double r1,complex z2,double r2)
     /* compute inversive distance between two eucl circles. Return 
negative value if overlap more than PI/2. Return default 1.0 if 
situation not yet covered. */
{
  double ab;

  if (r1>0 && r2>0)
    {
      ab=(z1.re-z2.re)*(z1.re-z2.re)+(z1.im-z2.im)*(z1.im-z2.im);
      return ((ab-(r1*r1+r2*r2))/(2.0*r1*r2));
    }
  /* need to handle: one plane, r=0, or one negative r */
  return 1.0;
} /* inv_dist */

int e_dilatation(double A,double B,double C,
		    double a,double b,double c,double *retval)
     /* compute dilatation of barycentric map between euclidean triangles
with edge lengths <A,B,C> in domain and <a,b,c> (corresponding) in range, 
return via retval. If one or both triangles are degenerate or there is
another error, return 0. */
{
  int act;
  double x,y,u,v,cos1,cos2,sin1,sin2;
  double AB2,ab2;

  *retval=1.0;
  if ((AB2=2.0*A*B)<OKERR || (ab2=2.0*a*b)<OKERR 
      || C<OKERR || c<OKERR
      || (A>=(B+C+OKERR) && B>=(A+C+OKERR) && C>=(A+B+OKERR))
      || (a>=(b+c+OKERR) && b>=(a+c+OKERR) && c>=(a+b+OKERR))
      ) return 0;

  cos1=(A*A+B*B-C*C)/AB2; /* cos(theta) */
  cos2=(a*a+b*b-c*c)/ab2;
  sin1=sqrt(1-cos1*cos1);
  sin2=sqrt(1-cos2*cos2);
  x=B*cos1;
  y=1.0/(B*sin1);
  u=b*cos2;
  v=b*sin2;
  if (!(act=affine_dilatation(a/A, (u-(a/A)*x)*y,(double) 0.0, v*y,retval)))
    return 0;
  return 1;
} /* e_dilatation */

int affine_dilatation(double a,double b,double c,double d,double *dil)
     /* Compute the real dilatation of affine transformation 
	for 2x2 real matrix <a b;c d>. return 0 on error. */
{
  double det,Q;

  *dil=1.0;
  if ((det=(a*d-b*c))<OKERR) return 0;
  Q=(a*a+b*b+c*c+d*d)/((2.0)*det);
  *dil=Q+sqrt(Q*Q-1.0);
  return 1;
} /* affine_dilatation */

complex quad_cross_ratio(double r1,double r2,double r3,double r4,
	     double e1,double e2,double e3,double e4,double e5)
     /* 4 euclidean circles are determining 2 faces; radii r1...r4
	counterclockwise, r1, r3 at ends of common edge; overlaps,
	e5 for common edge, ej for edge from r_j to r_{j+1}.
	Return cross ratio: (z1-z2)(z3-z4)/((z1-z4)(z3-z2)). */
{
  double L1,L2,L3,L4,L5,l5,l1c,l4c;
  complex z1,z2,z3,z4,w,num1,num2;

  /* compute edge lengths using radii/overlaps */
  L5=r1*r1+r3*r3+2*r1*r3*e5;
  l5=sqrt(L5);
  L1=r1*r1+r2*r2+2*r1*r2*e1;
  L2=r2*r2+r3*r3+2*r2*r3*e2;
  L3=r3*r3+r4*r4+2*r3*r4*e3;
  L4=r4*r4+r1*r1+2*r1*r4*e4;

  z1.re=z1.im=0.0;
  z3.re=l5;z3.im=0.0;
  l1c=(L1+L5-L2)/(2.0*l5);
  z2.re=l1c;z2.im=-sqrt(L1-l1c*l1c);
  l4c=(L4+L5-L3)/(2.0*l5);
  z4.re=l4c;z4.im=sqrt(L4-l4c*l4c);
  num1=cdiv(csub(z1,z2),csub(z1,z4));
  num2=cdiv(csub(z3,z4),csub(z3,z2));
  w=cmult(num1,num2);
  return w;
} /* quad_cross_ratio */

double invdist_length(double r1,double r2,double ovlap)
  /* return eucl length of edge with eucl radii r1, r2, and
     inv dist 'ovlap'. */
{
  return (sqrt(r1*r1+r2*r2+2.0*r1*r2*ovlap));
} /* invdist_length */
  
