#include "sa_types.h"
#include "sa_proto.h"

/* Parse and carry out options for 'fix' and 'Fix'; that is, laying 
   out a packing based on radii, checking combinatorics, etc.

Option:
   -a        default aims
   -c [dfsc] compute centers
             d   write diagnostics in /tmp/layout.log 
             f   use only 'well-placed' circles in layout 
             s   use drawing order only -- not average. 
             c x (with f) critical value x (double) 
   -d v      layout by drawing order, reporting location(s) of v.
   -F        redo everything
   -f {f..}  (must be last flag) redo facedraworder using face list
   -h vwn    drawing order by 'hex_walk' routine
   -K        redo combinatorics
   -r {f..}  recompute (don't draw) centers along given facelist   
   -s        recompute angle sums
   -l        suppress poorly placed circles (better to use -cf option above)
   -t        compute centers from 'tailored' drawing order; marked vertices
             will not (to extent possible) be used in drawing order.
   -T        same as -t, but routine will NOT use the marked vertices; it
	     will just give up once it has done all it can without them.
   -x        experimental routine.
*/

int fix_call(struct p_data *p,char *datastr)
/* for "Fix" command and fix calls */
{
  int i,count=0,hits,v,w,n,tick,d,first_face;
  int errflag=0,dflag=0,opt=2,flag=0,tflag=0;
  char ch,next[BUFSIZE];
  double crit=0;
  f_data *newfaces=NULL;
  struct Vertlist *facelist=NULL;
  char *nextptr,*endptr;

  nextptr=datastr;
  if (!p->status) return 0; /* no packing */
  while (grab_next(&nextptr,next) && next[0]=='-') /* flags set */
    {
      flag=1;
      switch(next[1])
	{
	case 'a': /* default aims */
	  {
	    set_aim_default(p);
	    count++;
	    break;
	  }
	case 'c': /* centers (with options) */
	  {
	    i=2;
	    while ((ch=next[i])!='\0')
	      {
		if (ch=='d') dflag=1;
		else if (ch=='f') errflag=1;
		else if (ch=='s') opt=1;
		else if (ch=='c' && sscanf(nextptr,"%lf",&crit));
		i++;
	      }
	    comp_pack_centers(p,errflag,dflag,opt,crit);
	    count++;
	    break;
	  }
	case 't': /* tailored drawing order; avoid using marked
		     vertices in layout ordering. */
	case 'T':
	  {
	    tflag=0;
	    if (next[1]=='T') tflag=1;
	    i=2;
	    while ((ch=next[i])!='\0')
	      {
		if (ch=='d') dflag=1;
		else if (ch=='f') errflag=1;
		else if (ch=='s') opt=1;
		else if (ch=='c' && sscanf(nextptr,"%lf",&crit));
		i++;
	      }
	    tick=0;
	    for (i=1;i<=p->nodecount;i++) if(p->packK_ptr[i].mark) tick++;
	    if (!tick)
	      {
		sprintf(msgbuf,"fix -[tT]: no vertices have been marked?");
		emsg();
	      }
	    else if ((newfaces=tailor_face_order(p,tflag,&first_face)))
	      {
		p->faces=newfaces;
		p->first_face=first_face;
		comp_pack_centers(p,errflag,dflag,opt,crit);
		count++;
	      }
	    break;
	  }
	case 'x': /* experimental */
	  {
	    i=2;
	    while ((ch=next[i])!='\0')
	      {
		if (ch=='d') dflag=1;
		else if (ch=='f') errflag=1;
		else if (ch=='s') opt=1;
		else if (ch=='c' && sscanf(nextptr,"%lf",&crit));
		i++;
	      }
	    tflag=0;
	    if ((newfaces=simple_layout(p,tflag,&first_face)))
	      {
		p->faces=newfaces;
		p->first_face=first_face;
		comp_pack_centers(p,errflag,dflag,opt,crit);
		count++;
	      }
	    break;
	  }
	case 'd': /* layout by drawing order, report, return */
	  {
	    v=grab_one_vert(p,&nextptr);
	    layout_report(p,v,1);
	    count++;
	    return count;
	  }			
	case 'e': /* given edgelist list; these are made 'poison'. */
	  {
	  	struct Edgelist *edgelist,*etrace;
	  	if ((edgelist=Node_pair_link(p,nextptr,&endptr,
			   &hits,&Vlist,&Elist,&Flist,&region,
			   pathlist,pathlength)) )
	      {
	      	for (i=1;i<=p->nodecount;i++) p->packK_ptr[i].util_flag=0;
		etrace=edgelist;
		while (etrace) {
			p->packK_ptr[etrace->v].util_flag=-1;
			p->packK_ptr[etrace->w].util_flag=-1;
			etrace=etrace->next;
		}
		count +=facedraworder(p,1);
		edge_free(&edgelist);
	      }
	    return count; /* this has to be last flag */
	  }
	case 'F': /* redo everything */
	  {
	    complex_count(p,1);
	    facedraworder(p,0);
	    fillcurves(p);
	    comp_pack_centers(p,errflag,dflag,opt,layout_th);
	    set_aim_default(p);
	    count++;
	    return count;
	  }
	case 'f': /* given face list */
	  {
	    if ( (facelist=Face_link_parse(p,nextptr,&endptr,
					   &hits,&Vlist,&Elist,&Flist,&region,
					   pathlist,pathlength)) )
	      {
		outer_poison(p,facelist);
		count +=facedraworder(p,1);
		vert_free(&facelist);
	      }
	    return count; /* this has to be last flag */
	  }
	case 'h': /* drawing order via hex_walk routine */
	  {
	    if ( sscanf(nextptr,"%d %d %d",&v,&w,&n)!=3
		 || !(facelist=try_hex_pg(p,v,w,n,&d,&d,&d,&d)) ) 
	      break;
	    if (!facelist) /* some error */
	      {
		sprintf(msgbuf,"hex_walk failed.");
		emsg();
	      }
	    else 
	      {
		outer_poison(p,facelist);
		count +=facedraworder(p,1);
	      }
	    vert_free(&facelist);
	    break;
	  }
	case 'K': /* combinatorics */
	  {
	    complex_count(p,0);
	    facedraworder(p,0);
	    count++;
	    break;
	  }
	case 'r': /* recompute centers along given facelist */
	  {
	    if ( (facelist=Face_link_parse(p,nextptr,&endptr,
					   &hits,&Vlist,&Elist,&Flist,&region,
					   pathlist,pathlength)) )
	      count += layout_facelist(p,&facelist,1,0);
	    break;
	  }
	case 's': /* angle sums */
	  {
	    fillcurves(p);
	    count++;
	    break;
	  }
	case 'l': /* suppress poorly placed circles 
		     (better to use -cf option above) */
	  {
	    count += comp_pack_centers(p,1,dflag,opt,layout_th);
	    if (count<(p->nodecount))
	      {
		sprintf(msgbuf,"CAUTION: using 'fix -l', %d "
			"circles were not reliably located "
			"and will generally not be drawn.",
			p->nodecount-count);
		msg();
	      }
	  }
	} /* end of switch */
    } /* end of while */
  if (flag) return count; /* done if options were specified */
  /* else, default actions */
  fillcurves(p);
  comp_pack_centers(p,errflag,dflag,opt,layout_th);
  count++;
  return count;
} /* fix_call */

