#include "sa_types.h"
#include "sa_proto.h"

/* For Jeff Berliss's work on trying to "flip" a torus complex
into all hexagonal. Based this on "legal_edge.cpp".

Now criteria are not on cross-ratio, but instead on radius,
degree, degree/radius average of contiguous petals, etc.

Generate, put in order, and return priority list of interior edges
based on criteria options

Criterion sets 'order_code':

   1    lowest degree, 
        highest avg radius of contig nghbs.

   2    lowest degree, 
        highest avg degree of contig nghbs.
   

Print out degree, nghbs rad, and nghbs degree.

*/


/* Here's a vector to hold info on edges. 
   want to know about contiguous neighbors and about their other
   common neighbor --- don't know if this will be used. */
struct Edge_Data{
  int v,u; // end vertices of an edge; parent v and petal u.
  int n; // next counterclockwise petal
  int dv,du,dn; // their degrees
  int o; // common nghb opposite to v
  int dg; // degrees of opposite
  double rv,ru,rn; // radii
  double ro; // radii of common nghbs
};
static int order_code;
struct Edge_Data *edge_data;
int *orderlist=NULL;

int priority_1(const void *e1,const void *e2);
int priority_2(const void *e1,const void *e2);

void qsort_flips(int left,int right);
void swap_edges(int i,int j);
int ck_for_swap(int down,int up);

struct Edgelist *flip_priority(struct p_data *p,char *datastr)
  /* this says to scan the packing, fill out a vector of 'edge_data',
determine the requested priority criteria, sort the vector by the
criteria, return edge list in priority order, and print for the
user desired into on the top 5 edges.

Sorting: Idea is to sort first to find the "pivot" vertex --- the
one you are trying to FIX. This is generally one which has degree 
too small. 

CAUTION: this means you want to flip continguous petals --- NOT 
flip an edge from the pivot itself.

*/
{
  int i,j,u,v,n,o,ix,edge_count=0,count=0,edge_ind;
  struct Edgelist *elist=NULL,*etrace;
  struct R_data *pR_ptr=p->packR_ptr;
  struct K_data *pK_ptr=p->packK_ptr;  
  
  edge_data=NULL;

  /* Check that this is closed surface, count the number of edges.*/
  if (!p->status || p->num_bdry_comp) {
    sprintf(msgbuf,"flip_priority: pack is not allow to have boundary.");
    emsg();
    return NULL;
  }

  /* Read off 'order_code': what priority do we want? 
     currently, limited to 1 or 2. */

  if (!sscanf(datastr,"%d",&order_code) || order_code<1
      || order_code>2) {
    sprintf(msgbuf,"flip_priority: need order_code (1 or 2).");
    emsg();
    return NULL;
  }

  /* Note: every edge will appear twice, so this is twice the
     true edge count. */
  for (v=1;v<=p->nodecount;v++) edge_count += pK_ptr[v].num;

  /* initialize Edge_Data space */
  edge_data=(struct Edge_Data *)
    calloc(edge_count+2,sizeof(struct Edge_Data));

  /* fill in data; supposedly all vertices are interior */

  edge_ind=1;
  for (v=1;v<=p->nodecount;v++)
    for (j=0;j<pK_ptr[v].num;j++) {
      u=pK_ptr[v].flower[j];
      n=pK_ptr[v].flower[j+1];
      /* find common nghb to <u,n> */
      ix=nghb(p,u,n);
      ix=((ix-1+pK_ptr[u].num)%(pK_ptr[u].num));
      o=pK_ptr[u].flower[ix]; 
      edge_data[edge_ind].v=v;
      edge_data[edge_ind].u=u;
      edge_data[edge_ind].n=n;
      edge_data[edge_ind].o=o;
      edge_data[edge_ind].dv=pK_ptr[v].num;
      edge_data[edge_ind].du=pK_ptr[u].num;
      edge_data[edge_ind].dn=pK_ptr[n].num;
      edge_data[edge_ind].dg=pK_ptr[o].num;
      edge_data[edge_ind].rv=pR_ptr[v].rad;
      edge_data[edge_ind].ru=pR_ptr[u].rad;
      edge_data[edge_ind].rn=pR_ptr[n].rad;
      edge_data[edge_ind].ro=pR_ptr[o].rad;
      edge_ind++;
    }

  /* SORT: User options set global 'order_code', which is used to
     choose the sorting routine. */

  if (order_code==1) 
    qsort(&(edge_data[1]),(size_t)edge_count,sizeof(struct Edge_Data),
	  priority_1);
  else if (order_code==2)
    qsort(&(edge_data[1]),(size_t)edge_count,sizeof(struct Edge_Data),
	  priority_2);

  /* LINKED_LIST: make up edge list of top 10 */

  elist=etrace=(struct Edgelist *)calloc(1,sizeof(struct Edgelist));
  count=edge_count;
  if (count>10) count=10;
  for (i=1;i<=count;i++) {
    etrace=etrace->next=
      (struct Edgelist *)calloc(1,sizeof(struct Edgelist));
    /* remember: list is in increasing order and the nghbs get flipped */
    etrace->v=edge_data[edge_count+1-i].u; 
    etrace->w=edge_data[edge_count+1-i].n;
  }
  etrace=elist; /* first spot is empty */
  elist=elist->next;
  free(etrace);
  
  /* Now put out info on the top five candidates. */

  if (count>5) count=5;
  sprintf(msgbuf,"priority_edges: criterion %d, top %d edges.",
	  order_code,count);
  msg();
      
  for (i=1;i<=count;i++) {  /* print info; use ratio of nghb radii to v */
    j=edge_count+1-i;
    sprintf(msgbuf,"  vert %d, deg %d: flip <%d,%d>. "
	    "rad-ratio=<%.2f,%.2f>, deg=<%d,%d>.\n"
	    "      opposite is %d, deg %d, rad-ratio %.2f.",
	    (v=edge_data[j].v),edge_data[j].dv,
	      edge_data[j].u,edge_data[j].n,
	      edge_data[j].ru/pR_ptr[v].rad,edge_data[j].rn/pR_ptr[v].rad,
	      edge_data[j].du,edge_data[j].dn,
	      edge_data[j].o,edge_data[j].dg,edge_data[j].ro/pR_ptr[v].rad);
      msg();
  }
  free(edge_data);
  edge_data=NULL;
  return elist;
} /* flip_priority */

/* ----------------- priority functions for various criteria ----- */

int priority_1(const void *e1,const void *e2)
  /* this function is called by the library 'qsort' routine.
     Return positive if e1 greater than e2, 0 if equal, negative
     if e1 less than e2. */
  /* e1/e2 are 'edge_data' info; this routine gives higher 
     priority to lowest degree and highest avg nghbs radii. */
{
  /* must cast incoming stuff to 'const' form */
  const struct Edge_Data *ed1=(struct Edge_Data *)e1;
  const struct Edge_Data *ed2=(struct Edge_Data *)e2;

  if (e1==e2) return 0;
  if (ed1->dv<ed2->dv) 
      return 1;
  if (ed1->dv==ed2->dv) { /* tied on degree */
      if ((ed1->ru+ed1->rn)>(ed2->ru+ed2->rn)) 
	return 1;
    }
  return -1;
} /* priority_1 */

int priority_2(const void *e1,const void *e2)
  /* this function is called by the library 'qsort' routine.
     Return positive if e1 greater than e2, 0 if equal, negative
     if e1 less than e2. */
  /* e1/e2 are 'edge_data' info; this routine gives higher 
     priority to lowest degree and highest avg nghbs degree. */
{
  /* must cast incoming stuff to 'const' form */
  const struct Edge_Data *ed1=(struct Edge_Data *)e1;
  const struct Edge_Data *ed2=(struct Edge_Data *)e2;

  if (e1==e2) return 0;
  if (ed1->dv<ed2->dv) 
      return 1;
  if (ed1->dv==ed2->dv) { /* tied on degree */
      if ((ed1->du+ed1->dn)>(ed2->du+ed2->dn)) 
	return 1;
    }
  return -1;
} /* priority_2 */
  
