#include "sa_types.h"
#include "sa_proto.h"
#include <stdio.h>

/* Compute radii of packing to meet angle_sum targets, specified as
"aims". A negative "aim" means that that circle should not have its
radius adjusted. aim of zero is possible only for boundary 
circles in hyp setting. 

All routines are iterative, inspired by routine suggested
by Thurston. This file includes recent algorithms, old reliable
versions, and experimental stuff being tried in spherical case.

For hyperbolic, use x-radii, see hyp_math.c. If h is infinite, 
store eucl radius with negative value for use in graphing its horocycle.

There currently is no packing algorithm for spherical geometry. Only
approach is to force packing into unit disc, do maximal packing
there and project back to sphere. Often, this involves a "puncture"
of the complex; missing circle is replaced as northern hemisphere
after max packing is projected back to sph. 

Overlap angles specified > Pi/2 can lead to incompatibilities. To
avoid domain errors, routines computing angles in such cases return
angle Pi. */

/* Added arguments are positional: if omitted, they take default values 
as given in the header. Namely,

passNo = PASS_FIRST | PASS_MIDDLE | PASS_LAST
cutp = NULL
state = NULL

This tells riffle to first allocate memory, initialize superstep
variables (PASS_FIRST), do specified number of iterations
(PASS_MIDDLE), and clean up (PASS_LAST).  The two null pointers tell
riffle not to store any of the state variables from this call to
riffle, so that superstepping will have to start over the next time
riffle is called.  I.e. the default values for these last three
parameters should make modified routines behave exactly like the
original routines.
*/

typedef struct old_h_riffle_state  {
    int aimnum;
    int *index;
    double accum,cut, recip;
}  old_h_riffle_state;

int old_h_riffle(struct p_data *p,int passes, int passNo, double *cutp, void **state)
     /* pre 1996 riffle routine (slow, but dependable) -- */
{
  int i, j;
  int dummy,flag;
  int count = 0;
  double verr, err;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;
  old_h_riffle_state *localstate;

  /*
   * These variables should be restored from the state variable.
   */

  int aimnum;
  int *index;
  double accum,cut, recip;

  pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;

  if(passNo & PASS_ALLOC)
    {
      index=(int *)malloc(5*(p->nodecount+1)*sizeof(int));
      if(state != NULL)
        {
          localstate = (old_h_riffle_state *) malloc(sizeof(old_h_riffle_state));
          localstate->index = index;
          *state = (void *) localstate;
        }
    }
  if(passNo & PASS_INIT)
    {
      if(!(passNo & PASS_ALLOC))
        {
          if(state == NULL)
            {
              fprintf(stderr, "error in old_h_riffle: "
		      "passNo specifies PASS_INIT but not PASS_ALLOC, "
		      "and state = NULL (aborting)\n");
              return -1;
            }
          localstate = (old_h_riffle_state *) *state;
          index = localstate->index;
        }
      aimnum = 0;
      accum=0;
      for (i=1;i<=p->nodecount;i++)
        {
          if (pK_ptr[i].bdry_flag &&
              pR_ptr[i].aim>=0  && pR_ptr[i].aim<.001)
            pR_ptr[i].rad=(-.2);
          else if (pR_ptr[i].aim>0)
            {
              if (pR_ptr[i].rad<=0 && pR_ptr[i].aim>.00001) 
                pR_ptr[i].rad=.01;
              index[aimnum]=i;
              aimnum++;
              err=pR_ptr[i].curv-pR_ptr[i].aim;
              accum += (err<0) ? (-err) : err;
            }
        }
      if (aimnum==0) {free(index);return count;}
      recip=.333333/aimnum;
      cut=accum*recip;
      if(state != NULL)
        {
          localstate->aimnum = aimnum;
          localstate->accum = accum;
          localstate->cut = cut;
          localstate->recip = recip;
        }
    }

  if(passNo & PASS_MIDDLE)
    {
      if(state != NULL)
        {
          localstate = (old_h_riffle_state *) *state;
          aimnum = localstate->aimnum;
          index = localstate->index;
          accum = localstate->accum;
          cut = localstate->cut;
          recip = localstate->recip;
        }
      while ((cut >toler && count<passes) || (count == 0 && state != NULL))
        {
          for (j=0;j<aimnum;j++)
            {
              i=index[j];
              h_anglesum_overlap(p,i,pR_ptr[i].rad,
                                 &pR_ptr[i].curv,&flag);
              verr=pR_ptr[i].curv-pR_ptr[i].aim;
              if (fabs(verr)>cut)
                pR_ptr[i].rad=
		  h_radcalc(p,i,pR_ptr[i].rad,pR_ptr[i].aim,&dummy);
            }
          accum=0;
          for (j=0;j<aimnum;j++)
            {
              i=index[j];
              err=pR_ptr[i].curv-pR_ptr[i].aim;
              accum += (err<0) ? (-err) : err;
            }
          cut=accum*recip;
    
          /* show activity */
          if (!(count % 10)) repack_activity_msg("o");
            
          count++;
        } /* end of while */
      if(state != NULL)
        {
          localstate->aimnum = aimnum;
          localstate->index = index;
          localstate->accum = accum;
          localstate->cut = cut;
          localstate->recip = recip;
        }
    }

  if(cutp != NULL)  *cutp = cut;

  if(passNo & PASS_LAST)
    {
      if(state != NULL)
        {
          localstate = (old_h_riffle_state *) *state;
          aimnum = localstate->aimnum;
          index = localstate->index;
          accum = localstate->accum;
          cut = localstate->cut;
          recip = localstate->recip;
          free(localstate);
        }
      free(index);
    }

  return count;
} /* old_h_riffle */

int h_riffle_vert(struct p_data *p,int i)
{
  int n=0,dummy,flag;
  double diff;
  struct R_data *pR_ptr;

  pR_ptr=p->packR_ptr;
  h_anglesum_overlap(p,i,pR_ptr[i].rad,&pR_ptr[i].curv,&flag);
  diff=pR_ptr[i].curv-pR_ptr[i].aim;
  while (n<20 && (diff>okerr || diff<(-okerr)) )
    /* limit number of attempts */
    {
      pR_ptr[i].rad=h_radcalc(p,i,
                              pR_ptr[i].rad,pR_ptr[i].aim,&dummy);
      h_anglesum_overlap(p,i,pR_ptr[i].rad,&pR_ptr[i].curv,&flag);
      diff=pR_ptr[i].curv-pR_ptr[i].aim;
      n++;
    }
  return n;
} /* h_riffle_vert */

typedef struct h_riffle_state  {
    int aimnum,*index,key;
    int maxBadCuts;
    int minBadCuts;
    int sumBadCuts;
    int cntBadCuts;
    int sct, fct;
    double m, c0;
    double ttoler;
    double *R1, *R2;
}  h_riffle_state;

int h_riffle(struct p_data *p,int passes, int passNo, double *cutp, void **state)

/* CRC - modified 8/5/02 from h_riffle
   uses uniform neighbor model (UNM) and two levels of super steps
   has safety measures to avoid any bad steps

   Anglesum calculations are in-line. Adjust 
   hyperbolic s-radii to meet curvature targets in 'aim'. 
   Neg. aim means that vertex is free - not subject to adjustment. */

/* based on new_e_riffle in structure (see it for an outline) */

{
  int i, j, k, j1, j2, N, flag;
  int count = 0;
  int numBadCuts = 0;
  double mm, c1, act, pred, factor, lambda, mmax;
  double r, r1, r2, r3, fbest, faim, del, bet;
  double fact=-1,fact0,denom;
  double ftol = 0.0005;                        /* Type 2 tolerance */
  double lmax, rat, twor;
  double m2,m3,sr,t1,t2,t3,tr,o1,o2,o3;
  struct R_data *pR_ptr;
  struct K_data *pK_ptr;
  h_riffle_state *localstate;

  /*
   * These variables should be restored from the state variable.
   */

  int aimnum,*index,key;
  int maxBadCuts;
  int minBadCuts;
  int sumBadCuts;
  int cntBadCuts;
  int sct, fct;
  double m, c0;
  double ttoler;
  double *R1, *R2;

  pR_ptr=p->packR_ptr;
  pK_ptr=p->packK_ptr;

  if(passNo & PASS_ALLOC)
    {
      index=(int *)malloc(5*(p->nodecount+1)*sizeof(int));
      R1 = (double *)malloc((p->nodecount+1)*sizeof(double));
      R2 = (double *)malloc((p->nodecount+1)*sizeof(double));
      if(state != NULL)
        {
          localstate = (h_riffle_state *) malloc(sizeof(h_riffle_state));
          localstate->index = index;
          localstate->R1 = R1;
          localstate->R2 = R2;
          *state = (void *) localstate;
        }
    }
  if(passNo & PASS_INIT)
    {
      if(!(passNo & PASS_ALLOC))
        {
          if(state == NULL)
            {
              fprintf(stderr, "error in h_riffle: passNo specifies "
		      "PASS_INIT but not PASS_ALLOC, and state = "
		      "NULL (aborting)\n");
              return -1;
            }
          localstate = (h_riffle_state *) *state;
          index = localstate->index;
          R1 = localstate->R1;
          R2 = localstate->R2;
        }
      aimnum = 0;
      key = 0;
      maxBadCuts = 0;
      minBadCuts = 0;
      sumBadCuts = 0;
      cntBadCuts = 0;
      /* list radii for adjustment, convert x_radii, store as s^2 */
      for (i=1;i<=p->nodecount;i++)
        {
          if (pK_ptr[i].bdry_flag && pR_ptr[i].aim>=0 
              && pR_ptr[i].aim<.001)
            pR_ptr[i].rad = (-.2);
          else if (pR_ptr[i].aim>0)
            {
              if (pR_ptr[i].rad<=0 && pR_ptr[i].aim>.00001)
                pR_ptr[i].rad = .01;
              index[aimnum]=i;
              aimnum++;
            }
          if (pR_ptr[i].rad>0)
            pR_ptr[i].rad = 1.0-pR_ptr[i].rad;
        }
      if (aimnum==0) 
        {free(index);free(R1);free(R2);return count;}
    
    /* set up parameters */
    
      ttoler = 3*aimnum*toler;              /* adjust tolerance */
      key = 1;                              /* initial superstep type */
      m = 1;                                /* Type 1 multiplier */
      sct = 1;                              /* Type 1 count */
      fct = 2;                              /* Type 2 minimum count */
    
    /* do one iteration to get started */  
      c0 = 0;                                   
      for (j=0;j<aimnum;j++) 
      {
            fbest = 0;
            i = index[j];
            if ((r = pR_ptr[i].rad)<0) r = 0;
            sr = sqrt(r);
            N = pK_ptr[i].num;
            if (!p->overlap_status) /* no overlaps */
            {                       
               twor = 2*r; 
               r2 = pR_ptr[pK_ptr[i].flower[0]].rad;
               m2 = (r2>0) ? (1-r2)/(1-r*r2) : (double) 1;
               for (k=1;k<=N;k++)   /* loop through petals */
               {
                 r3 = pR_ptr[pK_ptr[i].flower[k]].rad;
                 m3 = (r3>0) ? (1-r3)/(1-r*r3) : (double) 1;
                 fbest += acos(1-twor*m2*m3); /* angle calc */
                 m2 = m3;
               } 
            }
            else /* with overlaps, old routine */
            {
               j2 = pK_ptr[i].flower[0];
               if ((r2 = pR_ptr[j2].rad) < 0) r2=(double) 0;
               o2=pK_ptr[i].overlaps[0];
               for (k=1;k<=N;k++)
               {
                  r1=r2;
                  o1=o2;
                  j1=j2;
                  j2=pK_ptr[i].flower[k];
                  if ((r2 = pR_ptr[j2].rad) < 0) r2=(double) 0;
                  o2=pK_ptr[i].overlaps[k];
                  o3=pK_ptr[j1].overlaps[nghb(p,j1,j2)];
                  fbest += acos(h_cos_s_overlap(1-r,1-r1,1-r2,o3,o2,o1,&flag));
               }
            }
            faim = pR_ptr[i].aim;   /* get target sum */
            /* set up for model */
            denom = 1/(2*((double)N));
            del = sin(faim*denom);
            bet = sin(fbest*denom);
            r2 = (bet-sr)/(bet*r-sr);       /* reference radius */
            if (r2>0)                       /* calc new label */
            {
               t1 = 1 - r2;
               t2 = 2*del;
               t3 = t2/(sqrt(t1*t1+t2*t2*r2)+t1);
               r2 = t3*t3; 
            }
            else
               r2 = del*del;                /* use lower limit */
            pR_ptr[i].rad = r2;             /* store new label */ 
            pR_ptr[i].curv = fbest;         /* store new anglesum */
            fbest = fbest-faim;
            c0 += fbest*fbest;              /* accumulate error */
      }
      c0 = sqrt(c0);
      if(state != NULL)
        {
          localstate->aimnum = aimnum;
          localstate->key = key;
          localstate->maxBadCuts = maxBadCuts;
          localstate->minBadCuts = minBadCuts;
          localstate->sumBadCuts = sumBadCuts;
          localstate->cntBadCuts = cntBadCuts;
          localstate->sct = sct;
          localstate->fct = fct;
          localstate->m = m;
          localstate->c0 = c0;
          localstate->ttoler = ttoler;
        }
    }

/* End Iteration Loop */
  if(passNo & PASS_MIDDLE)
    {
      if(state != NULL)
        {
          localstate = (h_riffle_state *) *state;
          aimnum = localstate->aimnum;
          index = localstate->index;
          key = localstate->key;
          maxBadCuts = localstate->maxBadCuts;
          minBadCuts = localstate->minBadCuts;
          sumBadCuts = localstate->sumBadCuts;
          cntBadCuts = localstate->cntBadCuts;
          sct = localstate->sct;
          fct = localstate->fct;
          m = localstate->m;
          c0 = localstate->c0;
          ttoler = localstate->ttoler;
          R1 = localstate->R1;
          R2 = localstate->R2;
        }
	/* Begin Main Loop */
      while ((c0 >ttoler && count<passes) 
	     || (count == 0 && state != NULL))
        {
          for (i=1;i<=p->nodecount;i++) R1[i] = pR_ptr[i].rad;
          
          numBadCuts = 0;
          do  {   // Make sure factor < 1.0
              c1 = 0;
              for (j=0;j<aimnum;j++) 
              {
                 fbest = 0;
                 i = index[j];
                 if ((r = pR_ptr[i].rad)<0) r = 0;
                 sr = sqrt(r);
                 N = pK_ptr[i].num;
                 if (!p->overlap_status) /* no overlaps */
                 {                  
                    twor = 2*r; 
                    r2 = pR_ptr[pK_ptr[i].flower[0]].rad;
                    m2 = (r2>0) ? (1-r2)/(1-r*r2) : (double) 1;
                    for (k=1;k<=N;k++)      /* loop through petals */
                    {
                      r3 = pR_ptr[pK_ptr[i].flower[k]].rad;
                      m3 = (r3>0) ? (1-r3)/(1-r*r3) : (double) 1;
                      fbest += acos(1-twor*m2*m3); /* angle calc */
                      m2 = m3;
                    } 
                 }
                 else /* with overlaps, old routine */
                 {
                    j2 = pK_ptr[i].flower[0];
                    if ((r2 = pR_ptr[j2].rad) < 0) r2=(double) 0;
                    o2=pK_ptr[i].overlaps[0];
                    for (k=1;k<=N;k++)
                    {
                       r1=r2;
                       o1=o2;
                       j1=j2;
                       j2=pK_ptr[i].flower[k];
                       if ((r2 = pR_ptr[j2].rad) < 0) r2=(double) 0;
                       o2=pK_ptr[i].overlaps[k];
                       o3=pK_ptr[j1].overlaps[nghb(p,j1,j2)];
                       fbest += acos(h_cos_s_overlap(1-r,
				1-r1,1-r2,o3,o2,o1,&flag));
                    }
                 }
                 faim = pR_ptr[i].aim;      /* get target sum */
                 /* set up for model */
                 denom = 1/(2*((double)N));
                 del = sin(faim*denom);
                 bet = sin(fbest*denom);
                 r2 = (bet-sr)/(bet*r-sr);  /* reference radius */
                 if (r2>0)                  /* calc new label */
                 {
                    t1 = 1 - r2;
                    t2 = 2*del;
                    t3 = t2/(sqrt(t1*t1+t2*t2*r2)+t1);
                    r2 = t3*t3; 
                 }
                 else
                    r2 = del*del;           /* use lower limit */
                 pR_ptr[i].rad = r2;                /* store new label */ 
                 pR_ptr[i].curv = fbest;            /* store new anglesum */
                 fbest = fbest-faim;
                 c1 += fbest*fbest;         /* accumulate error */
              }
              c1 = sqrt(c1);
    
              factor = c1/c0;
              if (factor >= 1.0) {
                 c0 = c1;
                 key = 1;
                 numBadCuts++;
              }
              if(numBadCuts > MAX_ALLOWABLE_BAD_CUTS) goto ABORT;
          } while(factor >= 1.0);
    
          if(cntBadCuts == 0)  {
            maxBadCuts = numBadCuts;
            minBadCuts = numBadCuts;
            sumBadCuts = numBadCuts;
          }
          else  {
            maxBadCuts = (numBadCuts > maxBadCuts) ? numBadCuts : maxBadCuts;
            minBadCuts = (numBadCuts < minBadCuts) ? numBadCuts : minBadCuts;
            sumBadCuts += numBadCuts;
          }
          cntBadCuts++;
    
        /* superstep calculation */
    
          /* save values */
          for (i=1;i<=p->nodecount;i++) R2[i] = pR_ptr[i].rad;
    
          /* find maximum step can safely take */
          lmax = 10000;
          for (j=0;j<aimnum;j++)         /* find max step */
          {
              i = index[j];
              r = pR_ptr[i].rad;
              rat = r - R1[i];
              if (rat>0)                        
                 lmax=(lmax < (tr=((1-r)/rat))) ? lmax : tr; /* to keep R<1*/
              else if (rat < 0)
                 lmax = (lmax < (tr= (-r/rat))) ? lmax : tr; /* to keep R>0 */
           }
           lmax = lmax/2;
    
           /* do super step */
           if (key==1)              /*  type 1  SS */
           {
              lambda = m*factor;
              mmax = 0.75/(1-factor);               /* upper limit on m */
              m = (mmax < (mm=(1+0.8/(sct+1))*m)) ? mmax : mm;
            }
            else                     /*  type 2 SS */
            {
              if (sct>fct && fabs(factor-fact0)<ftol)  /* try SS-2 */
              {
                 lambda = factor/(1-factor);
                 sct = -1;
              }
              else
                 lambda = factor;                  /* do something */
            }
            lambda = (lambda>lmax) ? lmax : lambda;
    
            /* interpolate new labels */                 
            for (j=0;j<aimnum;j++)
            {
                i = index[j];
                pR_ptr[i].rad += lambda*(pR_ptr[i].rad-R1[i]);
            }
            sct++;
            fact0 = factor;
    
    /* end of superstep */
    
    /* do step/check superstep */
            c0 = 0;                             
            for (j=0;j<aimnum;j++) 
            {
               fbest = 0;
               i = index[j];
               if ((r = pR_ptr[i].rad)<0) r = 0;
               sr = sqrt(r);
               N = pK_ptr[i].num;
               if (!p->overlap_status) /* no overlaps */
                 {                  
                   twor = 2*r; 
                   r2 = pR_ptr[pK_ptr[i].flower[0]].rad;
                   m2 = (r2>0) ? (1-r2)/(1-r*r2) : (double) 1;
                   for (k=1;k<=N;k++)       /* loop through petals */
                     {
                       r3 = pR_ptr[pK_ptr[i].flower[k]].rad;
                       m3 = (r3>0) ? (1-r3)/(1-r*r3) : (double) 1;
                       fbest += acos(1-twor*m2*m3); /* angle calc */
                       m2 = m3;
                     } 
                 }
               else /* with overlaps, old routine */
               {
                  j2 = pK_ptr[i].flower[0];
                  if ((r2 = pR_ptr[j2].rad) < 0) r2= (double) 0;
                  o2=pK_ptr[i].overlaps[0];
                  for (k=1;k<=N;k++)
                  {
                     r1=r2;
                     o1=o2;
                     j1=j2;
                     j2=pK_ptr[i].flower[k];
                     if ((r2 = pR_ptr[j2].rad) < 0) r2= (double) 0;
                     o2=pK_ptr[i].overlaps[k];
                     o3=pK_ptr[j1].overlaps[nghb(p,j1,j2)];
                     fbest += acos(h_cos_s_overlap(1-r,1-r1,1-r2,
						   o3,o2,o1,&flag));
                  }
               }
               faim = pR_ptr[i].aim;        /* get target sum */
               /* set up for model */
               denom = 1/(2*((double)N));
               del = sin(faim*denom);
               bet = sin(fbest*denom);
               r2 = (bet-sr)/(bet*r-sr);    /* reference radius */
               if (r2>0)                    /* calc new label */
               {
                  t1 = 1 - r2;
                  t2 = 2*del;
                  t3 = t2/(sqrt(t1*t1+t2*t2*r2)+t1);
                  r2 = t3*t3; 
               }
               else
                  r2 = del*del;             /* use lower limit */
               pR_ptr[i].rad = r2;          /* store new label */ 
               pR_ptr[i].curv = fbest;      /* store new anglesum */
               fbest = fbest-faim;
               c0 += fbest*fbest;           /* accumulate error */
         }
         c0 = sqrt(c0);
    
    /* check results */
         pred = exp(lambda*log(factor));    /* predicted improvement */
         act = c0/c1;                   /* actual improvement */
         if (act<1)                     /* did some good */
         {
            if (act>pred)             /* not as good as expected: reset */
            {
               m = 1;
               sct = 0;
              if (key==1) key = 2;
            }                       /* implied else: accept result */
         }
         else                           /* reset to before superstep */
         {
            m = 1;
            sct =0;
            for (i=1;i<=p->nodecount;i++) pR_ptr[i].rad  = R2[i];
            c0 = c1;
            if (key==2) key = 1;
         }
    
          /* show activity */
          if (!(count % 10)) repack_activity_msg("h");
                    
          count++;
        } /* end of main while loop */
      if(state != NULL)
        {
          localstate->aimnum = aimnum;
          localstate->index = index;
          localstate->key = key;
          localstate->maxBadCuts = maxBadCuts;
          localstate->minBadCuts = minBadCuts;
          localstate->sumBadCuts = sumBadCuts;
          localstate->cntBadCuts = cntBadCuts;
          localstate->sct = sct;
          localstate->fct = fct;
          localstate->m = m;
          localstate->c0 = c0;
          localstate->ttoler = ttoler;
          localstate->R1 = R1;
          localstate->R2 = R2;
        }
    }

  if(passNo & PASS_LAST)
    {
      if(state != NULL)
        {
          localstate = (h_riffle_state *) *state;
          aimnum = localstate->aimnum;
          index = localstate->index;
          key = localstate->key;
          maxBadCuts = localstate->maxBadCuts;
          minBadCuts = localstate->minBadCuts;
          sumBadCuts = localstate->sumBadCuts;
          cntBadCuts = localstate->cntBadCuts;
          sct = localstate->sct;
          fct = localstate->fct;
          m = localstate->m;
          c0 = localstate->c0;
          ttoler = localstate->ttoler;
          R1 = localstate->R1;
          R2 = localstate->R2;
          free(localstate);
        }
      for (i=1;i<=p->nodecount;i++)               /* reset x_radii */
        if (pR_ptr[i].rad>0)
          pR_ptr[i].rad = 1.0-pR_ptr[i].rad;
      free(index);
      free(R1);
      free(R2);
      if(sumBadCuts > 0)  {
        sprintf(msgbuf, "\nreturning:\n\tmaxBadCuts was %d\n"
		"\tminBadCuts was %d\n"
		"\tavgBadCuts was %.8lf\n\tfinal cut value was %le\n",
		maxBadCuts, minBadCuts, sumBadCuts / (singledouble) 
		cntBadCuts,(singledouble) c0);
        msg();
      }
    }
  if(cutp != NULL)  *cutp = c0;
  return count;
 ABORT:
  free(index);
  free(R1);
  free(R2);
  sprintf(msgbuf,"Aborting repack: radius went negative.");
  emsg();
  if(cutp != NULL)  *cutp = c0;
  return 0;
} /* h_riffle */
 
