#include "sa_types.h"
#include "sa_proto.h"

/* This is designed to return one or more display commands for
circles or faces in pack q based on locations relative to pack p.
This is for use with the java "Map" features, wherein side-by-side
canvasses of pack p and q are displayed: the user clicks circles or 
faces on one and the corresponding ones are displayed in the other.

Issue of translation: domain p may have a 'vertex_map', but it may or may 
not be appropriate to translate. For now, trans_flag is not used. 
Rather: if p->vertex_map is null or if p->nodecount==q->nodecount, we 
assume NO TRANSLATION.

The flags are in 'datastr'. Start with -c (or -v) for circles 
(default) or -f for faces. Then expect -z {x} {y} to indicate 
location in the geometry of p. Other options may be added later; 
e.g., might give a vlist or flist of objects. Return count of objects found.
*/

int face_map_action(struct p_data *p,struct p_data *q,complex z,int trans_flag);
int circle_map_action(struct p_data *p,struct p_data *q,complex z,int trans_flag);

int map_activity(struct p_data *p,struct p_data *q,char *datastr)
{
  int count=0,cir_face=0,dum,trans_flag=0,got_z_flag=0;
  double x,y;
  complex z;
  char *nextpoint=datastr;
  struct Vertlist *list=NULL,*hold_list=NULL;
  
  if (!p->status || !q->status) return 0;
  
  // currently, use translation only if p has vertex_map and
  //   nodecounts are different
  if (p->vertex_map && (p->nodecount!=q->nodecount)) trans_flag=1;
  
  // parse commands
  stripsp(nextpoint);
  if (!grab_next(&nextpoint,next)) return (openMap_call(p,q)); // default is open
  do {
    if (!strncmp(next,"-o",2)) {
      return (openMap_call(p,q));
    }
    if (!strncmp(next,"-x",2)) {
      return (closeMap_call());
    }
    else if (!strncmp(next,"-f",2)) { // defaults to circles
      cir_face=1;
    }
    else if (!strncmp(next,"-z",2)) {
      if (sscanf(nextpoint,"%lf %lf",&(z.re),&(z.im))!=2) return 0;
      if (p->hes>0) {
	double a;
	if ((a=(z.re*z.re+z.im*z.im))>=1.0) return 0; // not on sphere
	z=ss_view(p->screen,proj_vec_to_sph(sqrt(1.0-a),z.re,z.im),0,&dum);
      }
      got_z_flag=1;
    }
  } while (grab_next(&nextpoint,next));
  if (got_z_flag) {
    if (cir_face) return (face_map_action(p,q,z,trans_flag));
    return(circle_map_action(p,q,z,trans_flag));
  }
  return 0;
} /* map_activity */

int face_map_action(struct p_data *p,struct p_data *q,complex z,
int trans_flag) 
{
  int count=0;
  struct Vertlist *list=NULL;

  if (!(list=tri_search(p,z))) return 0;

  // show in p (domain)
  if (!list->next) { // only one in list? (save time)
    int f=list->v;
    sprintf(buf," -ff %d",f);
    sa_display_call(p,buf);
    // use vertices to find face in q (numbering may be different)
    if (trans_flag) {
      int nf=face_translation(p,f,q);
      if (nf>0) {
	sprintf(buf," -ff %d",nf);
	count +=sa_display_call(q,buf);
	sprintf(msgbuf," %d",f);
	smsg();
	return 1;
      }
      return 0;
    }
    
    // else, just check existence
    int nf;
    int a=p->faces[f].vert[0];
    int b=p->faces[f].vert[1];
    int c=p->faces[f].vert[2];
    if (a<=q->nodecount && b<=q->nodecount && c<=q->nodecount
	&& (nf=what_face(q,a,b,c))>0) {
      sprintf(buf," -ff %d",nf);
      count+=sa_display_call(q,buf);
      sprintf(msgbuf," %d",f);
      smsg();
      return 1;
    }
    return 0;
  }

  // more than one?
  struct Vertlist *hold_list=p->flist;
  p->flist=list;
  sa_display_call(p," -ff flist"); 
  p->flist=hold_list;
      
  // show in q (range)
  struct Vertlist *trace=list;
  int f,nf;
  if (trans_flag) {
    while (trace) {
      f=trace->v;
      nf=face_translation(p,f,q);
      if (nf>0) {
	sprintf(buf," -ff %d",nf);
	count+=sa_display_call(q,buf);
	sprintf(msgbuf," %d",f);
	smsg();
      }
      trace=trace->next;
    }
    return count;
  }
  while (trace) { // just check existence
    f=trace->v;
    int a=p->faces[f].vert[0];
    int b=p->faces[f].vert[1];
    int c=p->faces[f].vert[2];
    if (a<=q->nodecount && b<=q->nodecount && c<=q->nodecount
	&& (nf=what_face(q,a,b,c))>0) {
      sprintf(buf," -ff %d",nf);
      count+=sa_display_call(q,buf);
      sprintf(msgbuf," %d",f);
      smsg();
    }
    trace=trace->next;
  }
  return count;
} /* face_map_action */
 
int circle_map_action(struct p_data *p,struct p_data *q,complex z,int trans_flag)
{
  int count=0;
  struct Vertlist *list=NULL;

  if (!(list=cir_search(p,z))) return 0;

  // show in p (domain)
  int nv;
  if (!list->next) { // shortcut: only one in list? 
    sprintf(buf," -cf %d",list->v);
    sa_display_call(p,buf);
    if (trans_flag) {
      if ((nv=vertex_translation(p->vertex_map,list->v))>0 
	  && nv<=q->nodecount) {
	sprintf(buf," -cf %d",nv);
	sa_display_call(q,buf);
	sprintf(msgbuf," %d",list->v);
	smsg();
	return 1;
      }
      return 0;
    }
    // else just check legality
    if (list->v<=q->nodecount) {
	sprintf(buf," -cf %d",list->v);
	sa_display_call(q,buf);
	sprintf(msgbuf," %d",list->v);
	smsg();
	return 1;
      }
    return 0;
  }

  // else, have list to process
  struct Vertlist *hold_list=p->vlist;
  p->vlist=list;
  sa_display_call(p," -cf vlist"); 
  // put indices in scratch window (even though some might
  //   not be found in q)
  Vertlist *trace=p->vlist;
  while (trace) {
    sprintf(msgbuf," %d",trace->v);
    smsg();
    trace=trace->next;
  }
  p->vlist=hold_list;
  
  // show in q (range)
  hold_list=q->vlist;
  q->vlist=list;
  if (trans_flag) {
    if (trans_rectify_vlist(p,q,&(q->vlist)))
      count=sa_display_call(q," -cf vlist"); 
  }
  else if (rectify_vlist(q,&(q->vlist)))
    count=sa_display_call(q," -cf vlist"); 
  vert_free(&(q->vlist));
  q->vlist=hold_list;
  return count;
} /* circle_map_activity */
