#include "sa_types.h"
#include "sa_proto.h"

/* given a side_pairing index n, record the mobius transform for
the side-pairing; this map carries the mate to side n to match 
side n iself. */

int pair_mobius(struct p_data *p,int n)
{
  int length=0,e_indx,m_indx,err_flag=0,N,inx;
  complex a,b,A,B,c,C;
  struct RedList *edge,*mate,*nextface,*endface,*trace,*midface;

  while (p->edge_pair[length+1].edge)
    length++;
  if (n<1 || n>length) return 0;
  p->edge_pair[n].mob=ident_mob();
  p->edge_pair[n].mob_err=0.0;
  if (!(mate=p->edge_pair[n].mate)) /* no edge-pairing recorded */
      return 0;

  edge=p->edge_pair[n].edge;
  e_indx=p->edge_pair[n].edge_indx;
  m_indx=p->edge_pair[n].mate_indx;

  if (!find_paired_centers(p,edge,e_indx,&A,&a)
      || !find_paired_centers(p,mate,m_indx,&b,&B))
    return 0; /* didn't find valid data */
  
  /* now, have everything for mobius transformation. (The mobius
     for mate is set separately (should be inverse to this mobius)). */

  if (p->hes<0) /* hyp */
    {
      if (cAbs(a)>=MOD1 && cAbs(b)>=MOD1 && cAbs(A)>=MOD1 && cAbs(B)>=MOD1)
	/* have another real degree of freedom, so we go to a face
	   midway along this paired edge chain and match its centers.*/
	{
	  /* find the last redface of this edge-pair segment; go one
	     face downstream from mate, jump across. */
	  inx=(m_indx+2)%3;
	  if (!(nextface=next_red_edge(p,1,mate,inx,&inx))
	      || !(endface=nextface->cross[inx])) 
	    return 0; /* some error in data */
	  length=1;
	  inx=(e_indx+2)%3;
	  trace=next_red_edge(p,1,edge,inx,&inx);
	  while (trace && trace!= endface 
		 && length<10000) 
	    {
	      length++;
	      trace=next_red_edge(p,1,trace,inx,&inx);
	    }
	  if (length==10000 || trace->next==edge) 
	    /* error, went all the way around? */
	    return 0;
	  if (length==1)
	    /* only one edge in this paired segment? Don't get proper 
	       parameter, so return error. Seems like this could happen 
	       legitimately, I just don't know what to do. */
	    return 0;
	  N=(int) length/2;
	  length=1;
	  inx=(e_indx+2)%3;
	  midface=next_red_edge(p,1,edge,inx,&inx);
	  /* march to approx midpoint of segment */
	  while (midface && length<N)
	    {
	      length++;
	      midface=next_red_edge(p,1,midface,inx,&inx);
	    }
	  if (!find_paired_centers(p,midface,(inx+1)%3,&C,&c)
	      || cAbs(c)>=MOD1 || cAbs(C)>=MOD1)
	    return 0; /* didn't get matching centers or one/both these
			 centers were too close to unit circle. */
	  p->edge_pair[n].mob=trans_abAB(a,b,A,B,c,C,&err_flag);
	}
      else if (cAbs(a)>=MOD1 || cAbs(b)>=MOD1 
	       || cAbs(A)>=MOD1 || cAbs(B)>=MOD1)
	return 0; /* fixup: don't have routine to handle some of points 
		     "on" unit circle and others off. Also,
		     may have too much error for either trans_abAB
		     or auto_abAB. */
      else p->edge_pair[n].mob=
	     auto_abAB(a,b,A,B,&(p->edge_pair[n].mob_err),&err_flag);
    }
  else if (p->hes==0) /* eucl */
    p->edge_pair[n].mob=affine_mob(a,b,A,B,&err_flag);
  else p->edge_pair[n].mob=ident_mob(); /* sph not implemented */
  if (err_flag) return 0; /* errors in creating the mobius maps should yield
			     the identity mobius and set err_flag. */ 
  return 1;
} /* pair_mobius */

int find_paired_centers(struct p_data *p,struct RedList *redface,
			int e_indx,complex *c1,complex *c2)
     /* Given spot in the redchain and e_indx pointing to one of its
vertices, find center c1 for that circle and center c2 for same circle
but from opposite side of a side-paired edge. e_indx must point to 
vert for which redface (or a predecessor, if blue) keeps center info.
(Key example: redface=edge and e_indx=edge_indx from EdgePair structure.)
Return 0 on failure. */

/* There is some ambiguity here: this circle might be plotted many
   times, but using local cross info, we'll have at most two options. 
   flag=0: use paired edge opposite to red edge beginning at e_indx vert.
   flag=1; use paired edge opposite to previous red edge.
   Else return 0. */
{
  int bflag=0,indx,vert,flag=0,n_indx;
  struct RedList *crossface,*nextface,*prevface;

  vert=p->faces[redface->face].vert[e_indx]; /* vertex of interest */

  /* find first center and crossface; several situations to consider */

  if (e_indx==redface->v_flag) /* redface is responsible for this vert */
    {
      *c1=redface->center; 
      /* need to find who on the other side is responsible */
      nextface=next_red_edge(p,1,redface,(e_indx+2)%3,&n_indx);
      if (!(crossface=nextface->cross[n_indx])) 
	{
	  if (!(crossface=redface->cross[(e_indx+2)%3]))
	    return 0;
	  flag=1;
	}
    }
  else if ((redface->next->face==redface->prev->face)
	   && (e_indx==(redface->v_flag+1)%3)
	   && (prevface=back_thru_blue(p,redface))) /* blue */
    {
      *c1=prevface->center;
      indx=(e_indx+2)%3;
      nextface=next_red_edge(p,1,redface,indx,&indx);
      if (!(crossface=nextface->cross[indx]))
	{
	  if (!(crossface=redface->cross[e_indx]))
	    return 0;
	  flag=1;
	}
    }
  else return 0;

  /* now have crossface and flag; find who's responsible for vert. */
  
  if (crossface->prev->face==crossface->next->face) bflag=1;
  /* crossface is blue */
  indx=find_index(p,crossface->face,vert);
  if (!flag) /* vert is at END of relevant paired edge */
    {
      if (indx==crossface->v_flag)
	{
	  *c2=crossface->center;
	  return 1;
	}
      if (bflag && indx==(crossface->v_flag+1)%3
	  && (prevface=back_thru_blue(p,crossface)) )
	{
	  *c2=prevface->center;
	  return 1;
	}
      return 0; /* error */
    }
  /* vert is at BEGINNING of relevant paired edge */
  if (bflag && indx==crossface->v_flag)
    {
      *c2=crossface->center;
      return 1;
    }
  if (indx==(crossface->v_flag+2)%3)
    {
      nextface=back_red_face(p,crossface,vert);
      *c2=nextface->center;
      return 1;
    }
  return 0; /* error */
} /* find_paired_centers */
      

  




