#include "sa_types.h"
#include "sa_proto.h"

/* Alternative face drawing orders for simply connected complexes.
The aim is to avoid some layout problems by, e.g., avoiding use of
3/4-degree vertices and/or vertices having extremely small radii.
 
Marked vertices are the ones to be avoided: that is, lay out
all others first. If flag=0, then when you run out of other faces,
you place one using the marked vertices, then again try to lay
out others until you run into another wall, etc. If flag=1, then
never use the marked vertices --- just stop when all possible are
laid out without it.

We will keep a linked list of vertices whose faces we should 
search through for new vertices to place at a given stage, adding 
to the end of the list as we progress; when that's used up, we try 
(if flag=0) some that were skipped, creating a new list to search 
through. Cycle through this process until all circles are 
eventually placed or (eg. if flag=1) you give up.

We will return a whole new "faces" data structure and first_face;
it is the responsibility of the calling routine to use the new
stuff, throw out the redfaces, empty p->rwb_flags, allocate
edge_pair, set first_face, manage the new f_data, etc. 
This routine empties and uses vertex 'util_flag'. */

f_data  *tailor_face_order(struct p_data *p,int flag,int *first_face)
     /* create new f_data */
{
  int i,j,alpha,num,vert,face,v1,v2,f,lastface,indx,v,w,fstop,way;
  int nbr=0,hits=0,vertcount=0,org_flag=0,pickup=0,*forg,nnum;
  struct Vertlist *priortrace,*trace,*endmain,*endnew,*ntrace;
  struct Vertlist *newlist=NULL,*mainlist=NULL;
  f_data *newfaces=NULL;
  struct K_data *pK_ptr=p->packK_ptr;

  if ((p->euler!=1 && p->euler!=2) || p->genus!=0 || !p->faces)
    return NULL; /* not simply connected */
  if (!face_org && !(org_flag=catalog_faces(p))) /* catalog failure */
    return NULL;
  /* create "newfaces" space, copy most stuff over. */
  newfaces=(f_data *)calloc((size_t)(p->facecount+1),sizeof(f_data));
  for (j=1;j<=p->facecount;j++) 
    {
      newfaces[j]=p->faces[j];
      newfaces[j].plot_flag=
	newfaces[j].next_face=newfaces[j].index_flag=0;
    }
    
  /* initialize */
  mainlist=trace=priortrace=endmain=
    (struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
  newlist=ntrace=endnew=
    (struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
  for (i=1;i<=p->nodecount;i++) pK_ptr[i].util_flag=pK_ptr[i].plot_flag=0;

  /* start with alpha (regardless of its situation) and contiguous
     neighbors (unmarked, if possible). */
  alpha=p->alpha;
  v=pK_ptr[alpha].flower[0];
  w=pK_ptr[alpha].flower[1];
  if (pK_ptr[v].mark || pK_ptr[w].mark)
    {
      for (i=0;((i<pK_ptr[alpha].num) && !nbr);i++)
	if (!pK_ptr[(v=pK_ptr[alpha].flower[i])].mark
	    && !pK_ptr[(w=pK_ptr[alpha].flower[i+1])].mark)
	  nbr=i;
    }
  pK_ptr[alpha].plot_flag=pK_ptr[v].plot_flag=pK_ptr[w].plot_flag=1;
  for (i=0;i<=pK_ptr[alpha].num;i++)
    pK_ptr[pK_ptr[alpha].flower[i]].util_flag++;
  for (i=0;i<=pK_ptr[v].num;i++)
    pK_ptr[pK_ptr[v].flower[i]].util_flag++;
  for (i=0;i<=pK_ptr[w].num;i++)
    pK_ptr[pK_ptr[w].flower[i]].util_flag++;
  vertcount=3;
  endmain->v=alpha;
  endmain=endmain->next=(struct Vertlist *)
    calloc((size_t)1,sizeof(struct Vertlist));
  endmain->v=v;
  endmain=endmain->next=(struct Vertlist *)
    calloc((size_t)1,sizeof(struct Vertlist));
  endmain->v=w;
  *first_face=lastface=*(face_org[alpha]+nbr+1);
  newfaces[lastface].index_flag=find_index(p,lastface,alpha);
  newfaces[lastface].plot_flag=1;
  for (i=1;i<=p->facecount;i++) newfaces[i].next_face=lastface;


  /* We keep two lists of vertices: 'mainlist' is the one we're
     going through. As long as mainlist isn't empty and we're
     getting hits of new vertices, build 'newlist' of newly added
     vertices. When through mainlist, we adjoin newlist to end,
     pass through mainlist again. If we get through mainlist and
     newlist is NULL, we pass through mainlist to add first 'marked'
     vertex, add it to end of mainlist, then go through mainlist
     again. Keeping track in util_flag of number of nghbs placed
     (depends on with int/bdry), plot_flag indicates faces placed
     (though others may have all vertices placed). Remove verts 
     from mainlist if all nghbs placed. */

  hits=1;
  while (mainlist && (pickup || hits)) 
    {
      hits=0;
      priortrace=trace=mainlist;

      /* Go through mainlist of verts. Add new verts to newlist as
	 encountered, remove from mainlist those we're finished
	 with. */
      while (trace)
	{
	  vert=trace->v;
	  /* already hit all neighbors of this vert? */
	  if (pK_ptr[vert].util_flag
	      ==(pK_ptr[vert].num+pK_ptr[vert].bdry_flag))
	    {
	      if (trace==mainlist) /* at beginning of list */
		{
		  priortrace=trace=mainlist->next;
		  free(mainlist);
		  mainlist=trace;
		}
	      else if (trace==endmain) /* at end of list */
		{
		  endmain=priortrace;
		  endmain->next=trace=NULL;
		}
	      else 
		{
		  priortrace->next=trace->next;
		  free(trace);
		  trace=priortrace->next;
		}
	    }
	  else if (!pK_ptr[vert].mark || pickup) /* process this vert */
	    {
	      num=pK_ptr[vert].num; /* num of faces */
	      nnum=num+pK_ptr[vert].bdry_flag; /* num of nghbs */
	      forg=face_org[vert];
	      /* go through the faces */
	      fstop=1;
	      while (fstop && pK_ptr[vert].util_flag!=nnum)
		{
		  fstop=0;
		  for (f=1;(f<=num && pK_ptr[vert].util_flag!=nnum);f++) 
		    if (newfaces[(face=*(forg+f))].plot_flag<=0)
		    {
		      indx=find_index(p,face,vert);
		      v1=newfaces[face].vert[(indx+1)%3];
		      v2=newfaces[face].vert[(indx+2)%3];
		      /* can we place either v1 or v2? */
		      if (pK_ptr[v2].plot_flag>0 && pK_ptr[v1].plot_flag>0)
			newfaces[face].plot_flag=1;
		      else if (pK_ptr[v2].plot_flag<=0 
			       && pK_ptr[v1].plot_flag>0
			  && (pickup || !pK_ptr[v1].mark))
			{
			  newfaces[lastface].next_face=face;
			  lastface=face;
			  newfaces[face].index_flag=indx;
			  newfaces[face].plot_flag=1;
			  pK_ptr[v2].plot_flag=1;
			  way=pK_ptr[v2].num+pK_ptr[v2].bdry_flag;
			  for (i=0;i<way;i++)
			    pK_ptr[pK_ptr[v2].flower[i]].util_flag++;
			  hits++;
			  fstop++;
			  vertcount++;
			  if (pK_ptr[v2].util_flag<
			      (pK_ptr[v2].num+pK_ptr[v2].bdry_flag))
			    {
			      endnew=endnew->next=(struct Vertlist *)
				calloc((size_t)1,sizeof(struct Vertlist));
			      endnew->v=v2;
			    }
			}
		      else if (pK_ptr[v1].plot_flag<=0 
			       && pK_ptr[v2].plot_flag>0
			       && (pickup || !pK_ptr[v2].mark))
			{
			  newfaces[lastface].next_face=face;
			  lastface=face;
			  newfaces[face].index_flag=(indx+2)%3;
			  newfaces[face].plot_flag=1;
			  pK_ptr[v1].plot_flag=1;
			  way=pK_ptr[v1].num+pK_ptr[v1].bdry_flag;
			  for (i=0;i<way;i++)
			    pK_ptr[pK_ptr[v1].flower[i]].util_flag++;
			  hits++;
			  fstop++;
			  vertcount++;
			  if (pK_ptr[v1].util_flag<
			      (pK_ptr[v1].num+pK_ptr[v1].bdry_flag))
			    {
			      endnew=endnew->next=(struct Vertlist *)
				calloc((size_t)1,sizeof(struct Vertlist));
			      endnew->v=v1;
			    }
			}
		      if (pickup && hits) /* kick out of for and while's */
			{
			  f=num+1;
			  fstop=0;
			  trace=NULL;
			}
		    } /* end of for f */
		} /* end of face while */
	      if (trace)
		{
		  priortrace=trace;
		  trace=trace->next;
		}
	    } /* end of else processing */
	  else
	    {
	      priortrace=trace;
	      trace=trace->next;
	    }
	} /* end of trace while */
      if (hits)
	{
	  if (newlist->next) /* add newlist to end of mainlist */
	    {
	      endmain->next=newlist->next;
	      endmain=endnew;
	      newlist->next=NULL;
	      endnew=newlist;
	    }
	  pickup=0;
	}
      else if (!flag && !pickup) pickup=1; /* pickup=1 means to
					      use a marked vert;
					      we resort to this
					      only when flag=0. */
      else pickup=0; /* no hits among marked/unmarked; quit. */
    } /* end of mainlist while */
    
  /* clean up and leave */
  if (mainlist) vert_free(&mainlist);
  if (vertcount!=p->nodecount)
    sprintf(msgbuf,"tailor_face_order: only placed %d of %d vertices.",
	    vertcount,p->nodecount);
  else sprintf(msgbuf,"tailor_face_order success.");
  msg();
  if (org_flag) destroy_catalog();
  if (vertcount) return newfaces;
  free(newfaces);
  return NULL;
} /* tailor_face_order */
	      

		      
		      

		  




