#include "sa_types.h"
#include "sa_proto.h"

/* Determine the welding map h(t) between two maximal packings. 
h(t):[0,2pi] is an orientation preserving homeomorphism associated 
with a homeomorphsm of the unit circle. (Conventions vary: could 
define on [0,1]; could make h orientation-reversing.)

The packings have corresponding bdry vertices, h(t) interpolates 
linearly. The output file is either matlab, postscript, or 'map' 
form (suitable for reading as a welding map by routine 'weld'
which is the reverse of this routine). 

out_flag: 0=postscript, 1=postscript+popup, 2=matlab, 3=map
*/

int weld_map(struct p_data *p,struct p_data *q,int v,int w,
	     char *filename,int out_flag)
  /* fixup: eventually want end verts on p and q so we can do
     a portion of the bdry. Need additional arguments. */
{
  int np,nq,n,last,wflag=0,count=0;
  double pmax=0.0,qmax=0.0,dist,*x=NULL,*y=NULL;
  double minx,miny,maxx,maxy,factorx,factory,p_arg0,q_arg0;
  char *endptr;
  struct Vertlist *trace;
  FILE *fp=NULL;

  if (!p->status || !q->status || p->hes>=0 || q->hes>=0
      || p->euler!=1 || p->genus || q->euler!=1 || q->genus) {
    sprintf(msgbuf,"weld_map: packings must be hyperbolic, top "
	    "discs.");
    emsg();
    return 0;
  }
  
  if (v<1 || v>p->nodecount || !p->packK_ptr[v].bdry_flag 
      || w<1 || w>q->nodecount || !q->packK_ptr[w].bdry_flag 
      || !strlen(filename)) {
    sprintf(msgbuf,"weld_map: given vertices "
	    "improper, or filename missing.");
    emsg();
    return 0;
  }
  p->bdry_starts[1]=v;
  q->bdry_starts[1]=w;
  sprintf(buf,"b");
  if (p->vlist) vert_free(&p->vlist);
  if (q->vlist) vert_free(&q->vlist);
  if (!(p->vlist=Node_link_parse(p,buf,&endptr,&np,
	       &Vlist,&Elist,&Flist,&region,pathlist,pathlength))
      || !(q->vlist=Node_link_parse(q,buf,&endptr,&nq,
	       &Vlist,&Elist,&Flist,&region,pathlist,pathlength))
      || np!=nq) {
    sprintf(msgbuf,"weld_map: edge counts don't agree.");
    emsg();
    return 0;
  }

  /* check if bdry centers "on" unit circle */

  trace=p->vlist;
  while (trace) {
    pmax=
      ((dist=fabs(1.0-cAbs(p->packR_ptr[trace->v].center)))>pmax) ?
      dist : pmax;
    trace=trace->next;
  }
  trace=q->vlist;
  while (trace) {
    qmax=
      ((dist=fabs(1.0-cAbs(q->packR_ptr[trace->v].center)))>qmax) ?
      dist : qmax;
    trace=trace->next;
  }
  
  if(pmax>.01 || qmax>.01) {
    sprintf(msgbuf,"weld_map: packings don't seem to be maximal.");
    emsg();
    return 0;
  }

  
  /* must reverse the order of the q_list */

  vert_free(&q->vlist);
  trace=q->vlist=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
  last=trace->v=w;
  while (trace->v!=w || !(wflag++)) {
    trace=trace->next=
      (struct Vertlist *)calloc(1,sizeof(struct Vertlist));
    trace->v=last=q->packK_ptr[last].flower[q->packK_ptr[last].num];
  }

  /* close up p vlist list */
  trace=p->vlist;
  while(trace->next) trace=trace->next;
  trace->next=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
  trace->next->v=p->vlist->v;

  /* normalize arguments in both p and q to start at zero */
  x=(double *)calloc((np+2),sizeof(double));
  x[0]=0.0;
  p_arg0=Arg(p->packR_ptr[v].center);
  count=1;
  trace=p->vlist->next;
  while(trace) {
    x[count]=Arg(p->packR_ptr[trace->v].center)-p_arg0;
    while (x[count]<x[count-1]) x[count] += 2.0*M_PI;
    count++;
    trace=trace->next;
  }
  
  /* reverse direction on q */
  y=(double *)calloc((np+2),sizeof(double));
  y[0]=0.0;
  q_arg0=Arg(q->packR_ptr[w].center);
  count=1;
  trace=q->vlist->next;
  while(trace) {
    y[count]=q_arg0-Arg(q->packR_ptr[trace->v].center);
    while (y[count]<y[count-1]) y[count] += 2.0*M_PI;
    count++;
    trace=trace->next;
  }
  if (!(fp=fopen(filename,"w"))) {
    sprintf(msgbuf,"weld_map_call: failed to open file %s.",filename);
    emsg();
    free(x);
    free(y);
    return 0;
  }

  /* now output depends on option */
  if (out_flag==3) { /* weld map h: [0,2pi]-->[0,2pi]; should be
			ready to call in with 'weld_map_call' */
    /* normalize first */
    minx=miny=1000000;
    maxx=maxy=-1000000;
    for (n=0;n<count;n++) {
      if (x[n]< minx) minx=x[n];
      if (y[n]< miny) miny=y[n];
      if (x[n]> maxx) maxx=x[n];
      if (y[n]> maxy) maxy=y[n];
    }
    factorx=2.0*M_PI/(maxx-minx);
    factory=2.0*M_PI/(maxy-miny);
    
    fprintf(fp,"PATH\n");
    for (n=0;n<count;n++) {
      fprintf(fp,"%.12f  %.12f\n",(x[n]-minx)*factorx,(y[n]-miny)*factory);
    }
    fprintf(fp,"END\n");
    sprintf(msgbuf,"weld_map: output in PATH form in file %s",filename);
    msg();
    fclose(fp);
  }
  else if (out_flag==2) { /* matlab form */
  
    /* preample and size */
    if (!(fp=fopen(filename,"w"))) {
      sprintf(msgbuf,"weld_map_call: failed to open file %s.",filename);
      emsg();
      free(x);
      free(y);
      return 0;
    }
    fprintf(fp,"%% matlab file for welding map fingerprint:\n"
	    "%%  x gives arguments in domain; y, arguments in range; "
	    "lists are closed.\n\n"
	    "N=%d\n",np+1);
    fclose(fp);
    
    /* put domain arguments in vector x */
    sprintf(buf,"x=[\n :: Varg \"\n\" :: vlist :: ]\n\n >> %s",filename);
    if (!output_call(p,buf)) {
      sprintf(msgbuf,"weld_map_call: failed to output 'x' values.");
      emsg();
      free(x);
      free(y);
      return 0;
    }
    /* put range arguments in vector y */
    sprintf(buf,"y=[\n :: Varg \"\n\" :: vlist :: ]\n\n >> %s",
	    filename);
    if (!output_call(q,buf)) {
      sprintf(msgbuf,"weld_map_call: failed to output 'y' values.");
      emsg();
      free(x);
      free(y);
      return 0;
    }
    sprintf(msgbuf,"weld_map: output in matlab form in file %s",filename);
    msg();
    fclose(fp);
  }
  else { /* postscript */
    if(graph_popup(x,y,np+1,filename,out_flag)) {
      sprintf(msgbuf,"weld_map: output graph in file %s",filename);
      msg();
    }
    else {
      sprintf(msgbuf,"weld_map: call to graph popup failed.");
      emsg();
      free(x);
      free(y);
      return 0;
    }
  }
  free(x);
  free(y);
    
  return np;
} /* weld_map */

int weld_map_call(struct p_data *p,struct p_data *q,char *datastr)
  /* parse the call to weld map; only one output option allowed */
{
  int v,w,out_flag=0;
  char filename[NAME_MAX],*nextpoint,*spot;

  nextpoint=datastr;
  /* default output is postscript */
  if ((spot=strstr(nextpoint,"-g "))) { /* with popup */
    out_flag=1;
    *spot=' ';
    *(spot+1)=' ';
    *(spot+2)=' ';
  }
  else if ((spot=strstr(nextpoint,"-w "))) { /* weld map form */
    out_flag=3;
    *spot=' ';
    *(spot+1)=' ';
    *(spot+2)=' ';
  }
  else if ((spot=strstr(nextpoint,"-m "))) { /* matlab output */
    out_flag=2;
    *spot=' ';
    *(spot+1)=' ';
    *(spot+2)=' ';
  }
  if (!(v=grab_one_vert(p,&nextpoint))|| !(w=grab_one_vert(q,&nextpoint))
      || !sscanf(nextpoint,"%s",filename)) {
    sprintf(msgbuf,"weld_map: usage: {p} {q} [-mwg] {v} {w} {filename}");
    emsg();
    return 0;
  }
  return (weld_map(p,q,v,w,filename,out_flag));
} /* weld_map_call */

