#include "sa_types.h"
#include "sa_proto.h"

/* This is based on Brock William's code, though conventions regarding
the maps have often been changed. 

Goal is to weld together two packings along their bdry (eventually,
one or several parts of bdrys). Two cases covered so far involve
full bdrys: 
1. two hyperbolic max packings using a welding map
2. two eucl packings using 
  * euclidean bdry edgelength (normalized to 1)
  * euclidean using harmonic measure (i.e., associate length on
    bdry of each pack with arclength on unit circle for its maximal
    packing)
 
After these welds, end up with a combinatorial sphere with 
weld-vertices in vlist and elist. Depending on options, can do call
the 'adjoin' here or call separately now that bdrys have the same
number of vertices.

Objects involved: 

* max packings p and q.  We always weld p onto q.  That is, we assume
  the map takes the boundary of p onto the boundary of q. Here p plays
  role of interior of unit disc, q exterior, so the prescription in
  the welding map means we move counterclockwise around bdry of p
  and clockwise around q (different convention from 'adjoin').

* welding map: real function h:[0,2pi]-->[0,2pi] (reflecting a 
  homeomorphism of the circle). Conventions could change: e.g. could
  map [0,1], h could be orientation-reversing, etc.

* welding list: Intermediate data which specifies how vertices on
  the two packings are matched/augmented. For now this is stored
  in /tmp/CirclePack_weldlist.b 
  Format:
    V v
    N v
    V n
    N n
    ...
    V v   (note: original two first verts should be identified, as
                 should be original  last two verts)

  First column, capitalized, describes bdry of p, the second, lower
  case, bdry of q. Letters V/v represent original verts, N/n represent 
  new added verts.
     
  So reading through the file, a 'V v' means an existing vertex of p
  should weld to an existing vertex of q. 'N v' means a new vertex
  should be added to p and welded to an existing vertex of q; conversely 
  for 'V n'.
     
Here are routines involved; some could be called for specific tasks,
but generally they are parts underlying a single processing call.

weld_map_to_list: this is the main workhorse routine. It reads a 
                  real-valued welding map from a file and stores
                  the welding list of V's and N's. 
readweldmap: read a welding map into a Pathlist variable (generally
             temporary). Format for data is
PATH
x1 y1
x2 y2
...
xn yn
END

weld: This actually walks around the boundaries and figures out how
to match/augment the vertices based on a given welding list.

add_between: utility to actually add necessary vertices, fix flowers, etc.

weld_using_map: function normally called to gather all these operations
                together. need to give it the filename for a welding map.
		It ends up adjusting p and q; p should end up as a
		comb. sphere with vlist giving the welded vertices. 
*/ 

int weld_map_to_list(struct p_data *p, struct p_data *q, 
		     int p_start_vert, int q_start_vert, 
		     int p_count, int q_count,
		     char *weldmapfile, char *weldlistfile,int opt_flag)
  /* Reads the mapfile and outputs a weldlist file containing instructions
     for matching/adding vertices.  */
{
  int i,j,n,pn,qn,path_length;
  int p_hits,q_hits,last,*p_vert=NULL,*q_vert=NULL,full_flag=1;
  double x1, x2, y1, y2,tolr,p_arg,q_arg;
  double *p_coord=NULL,*q_coord=NULL;
  double p_ang0,q_ang0,pmax=0.0,qmax=0.0,dist,factor=1.0;
  char *endptr;
  struct Pathlist *path_list,*ptrc,*ptrace,*qtrace;
  FILE *fp=NULL;

  if(!readweldmap(weldmapfile, &path_list, &path_length)) return(0);

  /* fixup: need to tune this cutoff for rounding off a vertex */
  tolr=0.1; 

  /* fixup: should do more error checking */
  if((!p->packK_ptr[p_start_vert].bdry_flag)
     || (!q->packK_ptr[q_start_vert].bdry_flag)) {
    path_free(&path_list);
    return(0);
  }
  if(!(fp=fopen(weldlistfile,"w"))) {
    sprintf(msgbuf,"Error: Could not open %s.",weldlistfile);
    emsg();
    path_free(&path_list);
    return(0);
  }

  /* Create appropriate bdry lists in p_vert/q_vert vectors. Count = -1 
     indicates full bdry; we also set this if conditions force it. */

  if (p_count>2 && q_count>2) full_flag=0; /* not full bdry, provisionally */

  /* use vlists only for counting */
  p->vlist=Node_link_parse(p,"b",&endptr,&p_hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength);
  vert_free(&(p->vlist));
  q->vlist=Node_link_parse(q,"b",&endptr,&q_hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength);
  vert_free(&(q->vlist));

  /* create space */
  if (full_flag || p_count>=p_hits || q_count>=q_hits) {
    /* Not enough edges, have to do all of both bdry's */
    full_flag=1;
    p_count=p_hits; 
    q_count=q_hits;
  }
  p_vert=(int *)calloc((p_count+1),sizeof(int));
  q_vert=(int *)calloc((q_count+1),sizeof(int));

  /* for full bdry, close with repeated first entry */
  p_vert[0]=last=p_start_vert;
  for (n=1;n<=p_count;n++) /* p list positive, counterclockwise */
    p_vert[n]=last=p->packK_ptr[last].flower[0];
  q_vert[0]=last=q_start_vert;
  for (n=1;n<=q_count;n++) /* Recall: q list must be clockwise. */
    q_vert[n]=last=q->packK_ptr[last].flower[q->packK_ptr[last].num];

  /* Use opt_flag here (fixup: will have more options eventually).
     default: two hyperbolic max packings --- check that they
              appear to be maximal, use arguments of center points 
	      (increasing arguments normalized to [0,1]).
     eucl: use bdry edge length --- check that both are eucl;
           both are normalized to [0,1]. */

  if (!(opt_flag & (0001 | 0002 | 0004))) { /* first bit not set */
   if (!p->hes && !q->hes) opt_flag |= 0001; 
  }

  if (!(opt_flag & (0001 | 0002 | 0004))) { /* first bit not set:
					       hyp max packings */

    /* check if bdry centers "on" unit circle */

    for (n=0;n<=p_count;n++)
      pmax=((dist=fabs(1.0-cAbs(p->packR_ptr[p_vert[n]].center)))>pmax) ?
	dist : pmax;
    for (n=0;n<=q_count;n++)
      qmax=((dist=fabs(1.0-cAbs(q->packR_ptr[q_vert[n]].center)))>qmax) ?
	dist : qmax;
    
    if(pmax>.01 || qmax>.01) {
      sprintf(msgbuf,"weld_map: packings don't seem to be maximal.");
      emsg();
      path_free(&path_list);
      free(p_vert);
      free(q_vert);
      return 0;
    }

    /* put coords in vectors, normalize each to [0,1]. */
    
    factor=0.5/M_PI;
    p_coord=(double *)calloc((p_count+1),sizeof(double));
    p_ang0=Arg(p->packR_ptr[p_vert[0]].center);
    p_coord[0]=0.0;
    for (n=1;n<=p_count;n++) {
      p_arg=Arg(p->packR_ptr[p_vert[n]].center);
      p_coord[n]=(p_arg-p_ang0)*factor;
      /* make arguments monitone */
      while (p_coord[n]<p_coord[n-1]) p_coord[n] += 1.0;
    }
    for (n=1;n<=p_count;n++)
      p_coord[n] /=p_coord[p_count];
    
    q_coord=(double *)calloc((q_count+2),sizeof(double));
    q_ang0=Arg(q->packR_ptr[q_vert[0]].center);
    q_coord[0]=0.0;
    for (n=1;n<=q_count;n++) {
      q_arg=Arg(q->packR_ptr[q_vert[n]].center);
      q_coord[n]=(q_ang0-q_arg)*factor;
      /* make arguments monitone */
      while (q_coord[n]<q_coord[n-1]) q_coord[n] += 1.0;
    }
    for (n=1;n<=q_count;n++)
      q_coord[n] /=q_coord[q_count];

  } /* end of opt_flag=0 setup */

  else if (opt_flag & 0001) { /* use euclidean edge lengths,
				 normalized to 1 */
    if (p->hes || q->hes) { /* not eucl */
      sprintf(msgbuf,"weld: are packings euclidean (for this option)?");
      emsg();
      free(p_vert);
      free(q_vert);
      free(p_coord);
      free(q_coord);
      return 0;
    }
    p_coord=(double *)calloc((p_count+2),sizeof(double));
    p_coord[0]=0.0;
    for (n=1;n<=p_count;n++) 
      p_coord[n] = p_coord[n-1]
	+p->packR_ptr[p_vert[n-1]].rad
	+p->packR_ptr[p_vert[n]].rad;
    for (j=1;j<=p_count;j++) p_coord[j] /= p_coord[p_count]; /* normalize */
    q_coord=(double *)calloc((q_count+2),sizeof(double));
    q_coord[0]=0.0;
    for (n=1;n<=q_count;n++) 
      q_coord[n] = q_coord[n-1]
	+q->packR_ptr[q_vert[n-1]].rad
	+q->packR_ptr[q_vert[n]].rad;
    for (j=1;j<=q_count;j++) q_coord[j] /= q_coord[q_count]; /* normalize */
  } /* end of eucl case */

  /* Want to transfer p_coords/q_coords to common interval [0,1] so
     they can be compared. For balance we do this in a neutral way,
     in essence replacing the path x-->y by x-->(x+y)/2. This should
     maintain monotonicity, but not favor p or q. We simultaneously
     do the linear interpolation. */

  /* first with p */
  ptrc=path_list;
  x1=ptrc->x;
  x2=ptrc->next->x;
  y1=ptrc->y;
  y2=ptrc->next->y;
  pn=1;
  while(pn<p_count) { 
    while(ptrc && (x2<p_coord[pn])) { /* arrange x1<p_coord<x2 */
      ptrc=ptrc->next;
      x1=x2;
      x2=ptrc->x;
      y1=y2;
      y2=ptrc->y;
    }  
    if (!ptrc) { /* ran out of function? */
      sprintf(msgbuf,"weld_map_to_list: problem: coords don't match map.");
      emsg();
      free(p_vert);free(q_vert);free(p_coord);free(q_coord);
      return 0;
    }
    p_coord[pn]= 
      ((y1 + (p_coord[pn]-x1)*((y2-y1)/(x2-x1)))+p_coord[pn])/2.0; 
    pn++;
  } 
  /* get the last p coord */
  ptrc=path_list;
  while (ptrc->next) ptrc=ptrc->next;
  p_coord[p_count]=(ptrc->x+ptrc->y)/2.0;

  /* same with q */
  ptrc=path_list;
  x1=ptrc->x;
  x2=ptrc->next->x;
  y1=ptrc->y;
  y2=ptrc->next->y;
  qn=1;
  while(qn<q_count) { 
    while(ptrc && (y2<q_coord[qn])) { /* arrange y1<q_coord<y2 */
      ptrc=ptrc->next;
      x1=x2;
      x2=ptrc->x;
      y1=y2;
      y2=ptrc->y;
    }  
    if (qn<(q_count-1) && !ptrc) { /* ran out of function? */
      sprintf(msgbuf,"weld_map_to_list: problem: coords don't match map.");
      emsg();
      free(p_vert);free(q_vert);free(p_coord);free(q_coord);
      return 0;
    }
    q_coord[qn]= 
      ((x1 + (q_coord[qn]-y1)*((x2-x1)/(y2-y1)))+q_coord[qn])/2.0; 
    qn++;
  } 
  /* get the last q coord */
  ptrc=path_list;
  while (ptrc->next) ptrc=ptrc->next;
  q_coord[q_count]=(ptrc->x+ptrc->y)/2.0;


  /* Now for the determination of new vertex creation */

  fprintf(fp,"V v \n");  /* Always weld p_start_vert to q_start_vert 
			    for now, might change the convention later */

  /* Heart of the algorithm: walk around bdry of p and q looking at
     locations of p_vert and q_vert. (We always weld p onto q.)

     If q_vert maps "past" p_vert, then we could add vertex to q so
     that p_vert has a vertex to glue onto.  If q_vert maps to a 
     location "before" p_vert, then we could add a vertex to p so
     that q_vert has a vertex to glue onto. 

     Since p_coord/q_coord values are now intermingled on same 
     interval [0,1], just have to tick off new/existing verts in order.
     Hard part is deciding when two are close enough to be identified.*/

  pn=qn=1;
  while(pn<=p_count && qn<=q_count) {
    /* fixup: very hard to decide how to parse this. In the theory you
       need to avoid "thin triangles", mainly occurring when new verts 
       generated are too close to existing ones. To avoid this, try to put
       interval of attraction around each vert to tell when new verts should
       be identified. Simple rule now: bracket each vert by "tolr" 
       proportion of distance to previous/next vert, if they intersect,
       identify those verts. (This can be preempted if other verts
       intervene.) Also have to take care in approaching the end verts.*/

    if (pn==p_count && qn==q_count) { /* end of both lists, identify */
      fprintf(fp,"V v \n");
      pn++;
      qn++;
      break;
    }
    if (pn==p_count) {  /* end of p, must add new to q */
      fprintf(fp,"N v \n");
      qn++;
    }
    else if (qn==q_count) { /* end of q, must add to p */
      fprintf(fp,"V n \n");
      pn++;
    }
    else if (p_coord[pn]<=q_coord[qn]) {
      if (p_coord[pn+1]<=q_coord[qn]
	  || p_coord[pn]<(q_coord[qn]-tolr*(q_coord[qn]-q_coord[qn-1]))) {
	fprintf(fp,"V n \n");
	pn++;
      }
      else {
	fprintf(fp,"V v \n");
	qn++;
	pn++;
      }
    }
    else {
      if (q_coord[qn+1]<=p_coord[pn]
	  || q_coord[qn]<(p_coord[pn]-tolr*(p_coord[pn]-p_coord[pn-1]))) {
	fprintf(fp,"N v \n");
	qn++;
      }
      else {
	fprintf(fp,"V v \n");
	qn++;
	pn++;
      }
    }
  }
  
  fclose(fp);
  path_free(&path_list);
  free(p_vert);free(q_vert);free(p_coord);free(q_coord);
  return 1;
} /* weld_map_to_list */

int readweldmap(char *weldmapfile,
		struct Pathlist **path_list,int *path_length)
  /* Read a welding map into a pathlist (although it is not a geometric
     pathlist). This should be a one-to-one increasing map of [a,b]
     to [c,d] which is normalized to map [0,1] onto [0,1]. Modeled after 
     the readpath() function.  Also returns number of points specified 
     in the mapfile. Return 0 on error. (Try to avoid global names 
     'pathlist', 'pathlength'.) */	
{
  int count=0,dum=0;
  char command[128];
  double xx, yy,xmin,xmax,ymin,ymax,xdiff,ydiff;
  struct Pathlist *trace;
  FILE *fp=NULL;
 
  if (strlen(weldmapfile)<2) { /* generate the default identity file */
    *path_length=0;
    (*path_list)=trace=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
    trace->x=trace->y=0.0;
    trace=trace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
    trace->x=trace->y=0.5;
    trace=trace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
    trace->x=trace->y=1.0;
    (*path_length)=3;
    return 3;
  }
  if (!(fp=fopen(weldmapfile,"r"))) {
    sprintf(msgbuf,"Error: Could not open file %s.",weldmapfile);
    emsg();return(0);
  }

  /* fixup: trying to allow reading of files whether or not they
     start with PATH, but it's not working right.*/

  while ( (fscanf(fp,"%127s",command)!=EOF) && (strcmp(command,"PATH")) 
	  && (strcmp(command,"END")) ) {
    dum++;
  }
    

  if (strncmp(command,"END",3) && fscanf(fp,"%lf %lf",&xx,&yy)!=2) {
    fclose(fp);
    return 0; 
  }
  *path_length=0;
  (*path_list)=trace=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
  trace=trace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
  trace->x=xx;
  trace->y=yy;
  count++;

  while ( fscanf(fp,"%lf %lf",&xx,&yy)==2 && count<2000000) {
    if( (xx<trace->x) || (yy<trace->y) ) {
      sprintf(msgbuf,"The map is not monotone increasing.");
      emsg();path_free(path_list); *path_list=NULL; *path_length=0;
      fclose(fp);
      return(0);
    }
    trace=trace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
    trace->x=xx;
    trace->y=yy;
    count++;
  }
  trace=*path_list;
  *path_list=(*path_list)->next;
  free(trace); /* first spot was empty */

  /* normalize to increasing map of [0,1] onto [0,1]. */
  trace=*path_list;
  xmin=trace->x;
  ymin=trace->y;
  while(trace && trace->next) trace=trace->next;
  xmax=trace->x;
  ymax=trace->y;
  xdiff=xmax-(*path_list)->x;
  ydiff=ymax-(*path_list)->y;
  (*path_list)->x=(*path_list)->y=0.0;
  trace=(*path_list)->next;
  while(trace) {
    trace->x=(trace->x-xmin)/xdiff;
    trace->y=(trace->y-ymin)/ydiff;
    trace=trace->next;
  }
  *path_length=count;
  fclose(fp);
  return count;
} /* readweldmap */

int weld(struct p_data *p,struct p_data *q,int v,int w,
	 char *weldlistfile,int opt_flag)
/* Go counterclockwise around p, clockwise around q, starting at v
   and w, resp, until pasting directions in file are all done.
   
opt_flag: 0010 ==> puncture q->alpha, do the adjoin operation; 
otherwise, just create the two packings (unpunctured). */

/* fixup: don't know other uses of opt_flag yet. */
{
  int ans=1,count=0,v_next,w_next,v_orig=v,w_orig=w;
  int *Oldnew=NULL,hits,full_flag=0;
  char command[64],*endptr;
  FILE *fp=NULL;

  if (!p->status || !q->status || p->locks
      || (v>p->nodecount) || (w>q->nodecount) || v<1 || w<1
      || !p->packK_ptr[v].bdry_flag || !q->packK_ptr[w].bdry_flag) {
    sprintf(msgbuf,"weld: improper data; e.g., verts %d and %d cannot "
	    "be used",v,w);
    emsg();
    return(0);
  }
  /* fixup: might save orig p and q in case this bombs */

  /* Augment p and q */
  if(!(fp=fopen(weldlistfile,"r"))
     || !fscanf(fp,"%63s",command) || strcmp(command,"V")
     || !fscanf(fp,"%63s",command) || strcmp(command,"v")) { 
    sprintf(msgbuf,"weld: couldn't open %s, or first line not 'V v'.",
	    weldlistfile);
    emsg();
    return(0);
  }
  v_next=p->packK_ptr[v].flower[0];
  w_next=q->packK_ptr[w].flower[q->packK_ptr[w].num];
  while( (fscanf(fp,"%63s",command)!=EOF) 
	 && strcmp(command,"END") ) {
    count=count+1;
    if(strcmp(command,"V")==0) { 
      v=v_next;
      v_next=p->packK_ptr[v].flower[0];
    }
    else if (strcmp(command,"N")==0) {
      if(!add_between(p,v,v_next)) 
	return (-1);
      v=p->packK_ptr[v].flower[0];
      /* v=newly added N vertex */
    }
    else {  /* extraneous stuff? */
      sprintf(msgbuf,"weld: weld file format problem - first part");
      emsg();
      return -1;
    }
    if((fscanf(fp,"%63s",command)==EOF) 
       || (strcmp(command,"END")==0) ) {
      sprintf(msgbuf,"weld: weld file format problem - second part");
      emsg();
      return -1;
    }
		
    if(strcmp(command,"v")==0) {

      w=w_next;
      w_next=q->packK_ptr[w].flower[q->packK_ptr[w].num];
    }
    else if (strcmp(command,"n")==0) {
      if (!add_between(q,w_next,w)) 
	return -1; 
      w=q->packK_ptr[w].flower[q->packK_ptr[w].num];
    }
    else {  /* extraneous stuff? */
      sprintf(msgbuf,"weld: weld file format problem - first part");
      emsg();
      return -1;
    }
  }
  complex_count(p,0);facedraworder(p,0);
  complex_count(q,0);facedraworder(q,0);

  if (v==p->packK_ptr[v_orig].flower[0])
    sprintf(buf,"b");
  else {
    v=p->packK_ptr[v].flower[p->packK_ptr[v].num];
    sprintf(buf,"b(%d %d)",v_orig,v);
  }
  p->vlist=Node_link_parse(p,buf,&endptr,&hits,
	   &Vlist,&Elist,&Flist,&region,pathlist,pathlength);
  p->elist=Node_pair_link(p,buf,&endptr,&hits,
	   &Vlist,&Elist,&Flist,&region,pathlist,pathlength);
  if (w==q->packK_ptr[w_orig].flower[q->packK_ptr[w_orig].num])
    sprintf(buf,"b");
  else {
    w=q->packK_ptr[w].flower[0];
    sprintf(buf,"b(%d %d)",w,w_orig);
  }
  q->vlist=Node_link_parse(q,buf,&endptr,&hits,
	   &Vlist,&Elist,&Flist,&region,pathlist,pathlength);
  q->elist=Node_pair_link(q,buf,&endptr,&hits,
	   &Vlist,&Elist,&Flist,&region,pathlist,pathlength);

  /* Adjoin p and q? */
  if (opt_flag & 0010) { /* bit set means to adjoin */
    /* standard procedure: puncture q at alpha vert, which
     will be associated with infinity after welding. */

    q->active_node=w_orig; /* this will be adjusted in the next calls */
    if (!swap_nodes(q,q->alpha,q->nodecount) 
	|| !puncture(q,q->nodecount)) {
      sprintf(msgbuf,"weld: failed to puncture pack q; continue to get "
	      "spherical result.");
      emsg();
    }
    complex_count(q,0);facedraworder(q,0);
    w_orig=q->active_node;
    q->bdry_starts[1]=w_orig;
    
    if ((ans=adjoin(p,q,v_orig,w_orig,count,&Oldnew))) {
      /* fixup: note that 'adjoin' pastes p clockwise, q counterclockwise;
	 no problem for full bdry, but with partial bdry's, have to do
	 something else. */
      /* fixup: may want to use Oldnew to define vertex_map of q */
      for (v=1;v<=p->nodecount;v++) p->packK_ptr[v].plot_flag=1;
    }
    if (Oldnew) free(Oldnew);
  }
  else {
    sprintf(msgbuf,"weld: two packs appear ready to adjoint:\n"
	    " designated bdry vertices are %d and %d, resp.,\n"
	    " alpha vertex of outer pack is %d",v,w,q->alpha);
    msg();
  }
  return ans;
} /* weld */

int add_between(struct p_data *p,int v,int v_next)
/* Adds a new vert on boundary between v and v_next.
Assumes both v, v_next are on boundary and v_next=p->packK_ptr[v].flower[0].
Results in     v_next +++++ newvert +++++ v
		     ++	       +      ++
			++     +    ++
			   ++  u ++                              

Special cases: if v or v_next lies in just one face, must add interior
vert to which new vert is attached; put it on opposite edge. */
{
  int a,b,i,j,k,u,w,y,num,indx,new_v,*newflower;
  struct K_data *pK_ptr=p->packK_ptr;

  /* Basic checks */
  if (!pK_ptr[v].bdry_flag || !pK_ptr[v_next].bdry_flag) return 0;

  /* if v or v_next has just one face --- no int nghbs. Put new
     interior vertex in edge opposite v or v_next, then we can
     continue as normal. label this way, X=new_v;

                                             y
				       ++    +    ++
				    b ++++++ X ++++++ a
				       ++    +    ++
                                             w                 */
  if (pK_ptr[v].num==1 || pK_ptr[v_next].num==1) { /* special: one face */
    if (pK_ptr[v].num==1) {
      a=pK_ptr[v].flower[1];
      b=v_next;
      y=v;
    }
    else {
      a=v;
      b=pK_ptr[v_next].flower[0];
      y=v_next;
    }
    if (pK_ptr[a].num<=2 || pK_ptr[b].num<=2) return 0;
    w=pK_ptr[a].flower[2];

    if (p->nodecount>(p->sizelimit-5)) 
      alloc_pack_space(p,p->nodecount+100,1);
    pK_ptr=p->packK_ptr;
    new_v=p->nodecount+1;

    /* fix a */
    pK_ptr[a].flower[1]=new_v;

    /* fix b */
    pK_ptr[b].flower[pK_ptr[b].num-1];

    /* fix y (v or v_next) */
    newflower=(int *)calloc(4,sizeof(int));
    newflower[0]=b;
    newflower[1]=new_v;
    newflower[2]=a;
    free(pK_ptr[y].flower);
    pK_ptr[y].flower=newflower;
    pK_ptr[y].num=2;
    pK_ptr[y].plot_flag=1;

    /* fix new_v */
    newflower=(int *)calloc(5,sizeof(int));
    newflower[0]=newflower[4]=a;
    newflower[1]=y;
    newflower[2]=b;
    newflower[3]=w;
    pK_ptr[new_v].num=4;
    pK_ptr[new_v].bdry_flag=0;
    pK_ptr[new_v].flower=newflower;
    pK_ptr[new_v].plot_flag=1;
    pK_ptr[new_v].color=FG_COLOR;
    p->packR_ptr[new_v].rad=p->packR_ptr[v].rad;

    /* fix w */
    if ((indx=nghb(p,w,a))<0) return 0;
    newflower=(int *)calloc((pK_ptr[w].num+2),sizeof(int));
    for (j=0;j<=indx;j++) newflower[j]=pK_ptr[w].flower[j];
    for (j=indx+1;j<=pK_ptr[w].num;j++) newflower[j+1]=pK_ptr[w].flower[j];
    newflower[indx+1]=new_v;
    free(pK_ptr[w].flower);
    pK_ptr[w].flower=newflower;
    pK_ptr[w].num++;
    pK_ptr[w].plot_flag=1;

    p->nodecount++;
  }

  /* proceed with new vert between v and v_next */
  if (p->nodecount>(p->sizelimit-5)) 
    alloc_pack_space(p,p->nodecount+100,1);/* add to p */
  p->nodecount=p->nodecount+1;
  pK_ptr=p->packK_ptr;
  u=pK_ptr[v].flower[1];
	
  /* fix v and v_next */

  pK_ptr[v].flower[0]=p->nodecount;
  pK_ptr[v_next].flower[pK_ptr[v_next].num]=p->nodecount;
	
  /* fix new node flower */
  i=p->nodecount;

  pK_ptr[i].flower=(int *)calloc(4,sizeof(int));
  pK_ptr[i].flower[0]=v_next;
  pK_ptr[i].flower[1]=u;
  pK_ptr[i].flower[2]=v;
  pK_ptr[i].num=2;
  pK_ptr[i].bdry_flag=1;
  pK_ptr[i].color=FG_COLOR;
  p->packR_ptr[i].rad=p->packR_ptr[v].rad;
  p->packR_ptr[i].center.re=p->packR_ptr[i].center.im=0.0;

  /* fix u */

  newflower=(int *)calloc((size_t)(pK_ptr[u].num+2),sizeof(int));
  num=pK_ptr[u].num;
  pK_ptr[u].num++;
  indx=nghb(p,u,v);
  for (k=0;k<=indx;k++)
    newflower[k]=pK_ptr[u].flower[k];
  for (k=indx+1;k<=num;k++) 
    newflower[k+1]=pK_ptr[u].flower[k];
  newflower[indx+1]=p->nodecount;
  free(pK_ptr[u].flower);
  pK_ptr[u].flower=newflower;
  return (1);
} /* add_between */

int weld_using_map(struct p_data *p,struct p_data *q,
		   int p_start_vert,int q_start_vert,
		   int p_count,int q_count,char *weldmapfile,int opt_flag)
  /* Most comprehensive of procedures: does all steps. */
   
  /* fixup; Currently, weld list is assumed to be in 
       '/tmp/CirclePack_weldlist.b'; can change after debugging.
       (For now it might be helpful to be able to easily see the welding
       list it creates.)  Later, we might want to create a weld list
       data structure and just keep the list in memory.   */

  /* fixup: when q_count and p_count are -1, use whole bdry; that is
     all we're prepared for now. */
{
  int ans=0;
  char weldlistfile[NAME_MAX];

  strcpy(weldlistfile,"/tmp/CirclePack_weldlist.b");

  if (!weld_map_to_list(p,q,p_start_vert,q_start_vert,p_count,q_count,
			weldmapfile,weldlistfile,opt_flag))
      return 0;
  if ((ans=weld(p,q,p_start_vert,q_start_vert,weldlistfile,opt_flag))>0)
    {
      complex_count(p,0);complex_count(q,0);
      fillcurves(p);fillcurves(q);
      facedraworder(p,0);facedraworder(q,0);
      set_aim_default(p);set_aim_default(q);
      p->active_node=p->beta; q->active_node=q->beta;
      if (strlen(weldmapfile)<2) 
	sprintf(msgbuf,"weld_with_map: verts %d %d, weldmap=identity map;",
		p_start_vert,q_start_vert);
      else sprintf(msgbuf,"weld_with_map: verts %d %d, weldmapfile %s;",
		   p_start_vert,q_start_vert,weldmapfile);
      msg();
      return ans;
    }
  strcpy(msgbuf,"Something went wrong with 'weld' routine; packing(s) "
	 "may be corrupted.");
  emsg();
  return ans;
} /* weld_using_map */

