/*
 * ztopen:  Returns a FILE pointer for which all stdio library
 * reads/writes will be automatically de/compressed (unless
 * we are reading from an uncompressed file, in which case
 * normal reads are performed).
 *
 * The returned FILE pointer can be used as if it directly
 * references an uncompressed file, except that it cannot be
 * seeked.
 *
 * The returned FILE pointer can be closed using fclose().
 * The ztclose() routine has been converted into a macro that
 * call fclose().
 *
 * The automatic de/compression works by causing fp to reference
 * one end of a pipe; a de/compressor thread is using zlib to
 * transfer uncompressed data between the other end of the pipe
 * and the original FILE pointer.
 */

#include "sa_types.h"
#include "sa_proto.h"

#define ZT_MAXBUF            1024

#define MAGIC_LEN 2
#define MAGIC_NO  "\037\213"
#ifdef WIN32 // FA
#define S_ISREG(m)      (((m) & S_IFMT) == S_IFREG)
#endif


typedef struct zipperThreadArgs  {
    gzFile gzfp;
    int childd;
    char *mode;
}  zipperThreadArgs;

typedef struct zipFileInfo  {
    zipperThreadArgs *zta;
    int parentd;
}  zipFileInfo;

static int maxNoFiles;
static zipFileInfo **zipFileList = NULL;
static pthread_mutex_t *zipFileListLock = NULL;
static pthread_cond_t *zipFileListCond = NULL;

void cleanUpZipperThreads(void)

{
    int i;

    for(i = 0; i < maxNoFiles; i++)  {
        pthread_mutex_lock(zipFileListLock);
            if(zipFileList[i] != NULL)  {
                close(zipFileList[i]->parentd);
                pthread_cond_wait(zipFileListCond, zipFileListLock);
            }
        pthread_mutex_unlock(zipFileListLock);
    }
}   

void *zipperThread(void *v)

{
    zipperThreadArgs *zta = (zipperThreadArgs *) v;
    int len;
    int prefix;
    int nwritten;
    char buf[ZT_MAXBUF];

    if(strcmp(zta->mode, "r") == 0)  {
        while((len = gzread(zta->gzfp, buf, ZT_MAXBUF)) > 0)  {
            prefix = 0;
            while(prefix < len)  {
                nwritten = write(zta->childd, buf + prefix, len - prefix);
                if(nwritten < 0)  break;
                prefix += nwritten;
            }
            if(nwritten < 0)  break;
        }
    }
    else  {
        while((len = read(zta->childd, buf, ZT_MAXBUF)) > 0)  {
            prefix = 0;
            while(prefix < len)  {
                nwritten = gzwrite(zta->gzfp, buf + prefix, len - prefix);
                if(nwritten <= 0)  break;
                prefix += nwritten;
            }
            if(nwritten <= 0)  break;
        }
    }

    pthread_mutex_lock(zipFileListLock);
        free(zta->mode);
        gzclose(zta->gzfp);
        free(zipFileList[zta->childd]);
        zipFileList[zta->childd] = NULL;
        close(zta->childd);
        free(zta);
    pthread_cond_signal(zipFileListCond);
    pthread_mutex_unlock(zipFileListLock);

    pthread_exit(NULL);
	return NULL; // FA
}

FILE *ztopen(char *filename, char *mode)

{
    int d;

    if(strcmp(mode, "r") == 0)  {
        if((d = open(filename, O_RDONLY)) < 0)  return NULL;
    }
    else if(strcmp(mode, "w") == 0)  {
        if((d = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666)) < 0)  return NULL;
    }
    else  {
        return NULL;
    }

    return ztdopen(d, mode);
} /* ztopen */

FILE *ztdopen(int d, char *mode)

{
    int p[2];
    int childd, parentd;
// FA    void (*oldSigHandler)(int);
    gzFile gzfp;
    FILE *fp;
    int len, prefix;
    pthread_t tid;
    pthread_attr_t attr;
    zipperThreadArgs *zta;
    void *rv;
    unsigned char magicBuf[MAGIC_LEN];
    struct stat sb;

    if(zipFileList == NULL)  {
        maxNoFiles = CPget_max_open_files();
        if(maxNoFiles < 0)  return NULL;
        
        if((zipFileList = (zipFileInfo **) calloc(maxNoFiles, sizeof(zipFileInfo *))) == NULL ||
           (zipFileListLock = (pthread_mutex_t *) calloc(1, sizeof(pthread_mutex_t))) == NULL ||
           (zipFileListCond = (pthread_cond_t *) calloc(1, sizeof(pthread_cond_t)))    == NULL ||
            pthread_mutex_init(zipFileListLock, NULL) != 0 ||
            pthread_cond_init(zipFileListCond, NULL)  != 0)  {
            if(zipFileList != NULL)  free(zipFileList);
            if(zipFileListLock != NULL)  free(zipFileListLock);
            if(zipFileListCond != NULL)  free(zipFileListCond);
            return NULL;
        }

        atexit(cleanUpZipperThreads);
#ifndef WIN32
        /* Make sure broken pipes don't kill the program. */
        sighandler_t oldSigHandler = signal(SIGPIPE, SIG_IGN);
        if(oldSigHandler == SIG_ERR)  return NULL;
        /* But don't interfere with other signal handlers that may have been installed. */
        if(oldSigHandler != SIG_IGN && oldSigHandler != SIG_DFL)
            signal(SIGPIPE, oldSigHandler);
#endif
    }

    if(strcmp(mode, "r") == 0)  {
        if(fstat(d, &sb) < 0)  {
            close(d);
            return NULL;
        }
        if(S_ISREG(sb.st_mode))  {
            prefix = 0;
            while(prefix < MAGIC_LEN)  {
                len = read(d, magicBuf + prefix, MAGIC_LEN - prefix);
                prefix += len;
            }
            if(lseek(d, 0, SEEK_SET) < 0)  return NULL;
            if(memcmp(magicBuf, MAGIC_NO, MAGIC_LEN) != 0)  {
                close(d);
                return NULL;
            }
        }
    }
    else if(strcmp(mode, "w") == 0)  {
    }
    else  {
        return NULL;
    }
    if(pipe(p) < 0)  {
        return NULL;
    }

    if(strcmp(mode, "r") == 0)  {
        childd = p[1];
        parentd = p[0];
    }
    else  {
        childd = p[0];
        parentd = p[1];
    }

    gzfp = gzdopen(d, mode);
    if(gzfp == NULL)  {
        close(parentd);
        close(childd);
        close(d);
        return NULL;
    }

    if((fp = fdopen(parentd, mode)) == NULL)  {
        close(parentd);
        close(childd);
        gzclose(gzfp);
        return NULL;
    }

    zta = (zipperThreadArgs *) malloc(sizeof(zipperThreadArgs));
    zta->mode = strdup(mode);
    zta->childd = childd;
    zta->gzfp = gzfp;
    
    pthread_mutex_lock(zipFileListLock);
        if(zipFileList[childd] != NULL)  {
            fprintf(stderr, "unknown internal error (aborting)\n");
            exit(1);

        }
        zipFileList[childd] = (zipFileInfo *) malloc(sizeof(zipFileInfo));
        if(zipFileList[childd] == NULL)  {
            fclose(fp);
            return NULL;
        }
        zipFileList[childd]->zta = zta;
        zipFileList[childd]->parentd = fileno(fp);
    pthread_mutex_unlock(zipFileListLock);

    if(pthread_attr_init(&attr)                                     != 0 ||
       pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)  != 0 ||
       pthread_create(&tid, &attr, zipperThread, zta)               != 0)  {
        fclose(fp);
        close(childd);
        gzclose(gzfp);
        return NULL;
    }

    return fp;
} /* ztdopen */
