#include "sa_types.h"
#include "sa_proto.h"

/* See 'x_riffle' in 'h_riffle_stuff.cpp' for original code. 
Data is stored in persistent 'state' variables so later calls 
can restart the computation --- for use in MPI parallel code. 

Idea here was to improve angle sum computation by exploiting O3 
optimization. 'checkout' marks some spots for consideration. 
Should learn to profile to find hangups. */

typedef struct x_riffle_state  {
    int aimnum,*index,key;
    int maxBadCuts;
    int minBadCuts;
    int sumBadCuts;
    int cntBadCuts;
    int sct, fct;
    FILE *dbugfile;
    double m, c0;
    double ttoler;
    double lmax,*R1, *R2;
}  x_riffle_state;

/* (local) global variables: available to subroutines.  */
/* fixup: can we define these JUST for this file? */

int aimnum,*indx;
double switchtol,*rad,*pdel;
double locr[11],mval[11],rslt[11],sq_rslt[11];
struct p_data *p;
struct R_data *h_pR_ptr;
struct K_data *h_pK_ptr;
double h_rif_routine(int method_flag);

int h_riffle_O3(struct p_data *p_in,int passes,int passNo,
		double *cutp,void **state)
{
  /* Modified_h_riffle to utilize the x-radii. For hyp label h,
   s-radius is s = exp(-h) and x-radius is x = 1 - s^2 (i.e.,
   1 - exp(-2h)). Advantages: (1) handles small h accurately
   (2) for small h, x is approximately euclidean so
   can sometimes use the (cheaper) euclidean routines.
   Has all the super steps and safety measures of other
   riffles. Does not do overlaps!  (use h_riffle instead)

   Various optional methods were tried with O3 optimization. So far,
   method_flag 4 performs best on intel with -O3. Should try additional
   optimization flags.
 */

  int method_flag=4; 

  int i,j,k,N,numBadCuts,count=0;
  int dbg=0;        /* set this flag for "debug.out" file */
  double mm,tr,r,c0,c1,act, pred,lambda,mmax,factor,fact0,rat;
  double ftol = 0.0005;                        /* Type 2 tolerance */
  x_riffle_state *localstate;

  /* These (and other) variables are restored from state variable. */
  int key,sct, fct;
  int maxBadCuts,minBadCuts,sumBadCuts,cntBadCuts;
  FILE *dbugfile=NULL;
  double m,ttoler,lmax,*R1, *R2;

  /* set up local-global variables */
  switchtol=1e-3;
  p=p_in; 
  rad=pdel=NULL;
  h_pR_ptr=p->packR_ptr;
  h_pK_ptr=p->packK_ptr;

  /* allocate new tmp space for radii/aims */
  rad=(double *)calloc((p->nodecount+1),sizeof(double));
  pdel=(double *)calloc((p->nodecount+1),sizeof(double));
  for (i=1;i<=p->nodecount;i++) {
    rad[i]=h_pR_ptr[i].rad;
    pdel[i]=sin(h_pR_ptr[i].aim/(2*h_pK_ptr[i].num));
  }

  if(passNo & PASS_ALLOC) {
    indx=(int *)malloc(5*(p->nodecount+1)*sizeof(int));
    R1 = (double *)malloc((p->nodecount+1)*sizeof(double));
    R2 = (double *)malloc((p->nodecount+1)*sizeof(double));
    if(state != NULL) {
      localstate = (x_riffle_state *) malloc(sizeof(x_riffle_state));
      localstate->index = indx;
      localstate->R1 = R1;
      localstate->R2 = R2;
      *state = (void *) localstate;
    }
  } /* end of PASS_ALLOC case */

  if(passNo & PASS_INIT) {
    if(!(passNo & PASS_ALLOC)) {
      if(state == NULL) {
	fprintf(stderr, "error in h_riffle_O3: passNo "
		"specifies PASS_INIT but not PASS_ALLOC, "
		"and state = NULL (aborting)\n");
	return -1;
      }
      localstate = (x_riffle_state *) *state;
      indx = localstate->index;
      R1 = localstate->R1;
      R2 = localstate->R2;
    }
    maxBadCuts = 0;
    minBadCuts = 0;
    sumBadCuts = 0;
    cntBadCuts = 0;
    lmax = 0;
    aimnum = 0;
    key = 0;

    /* open file to receive status reports */
    if (dbg==1) dbugfile = fopen("debug.out","w"); 
      
    /* list radii to be adjusted */
    for (i=1;i<=p->nodecount;i++) {
      if (h_pK_ptr[i].bdry_flag && h_pR_ptr[i].aim>=0 
	  && h_pR_ptr[i].aim<.0001)
	rad[i] = (-.2);
      else if (h_pR_ptr[i].aim>0 && h_pK_ptr[i].num>1) {
	if (rad[i]<=0 && h_pR_ptr[i].aim>.00001)
	  rad[i]= .01;
	indx[aimnum++]=i;
      }
    }
    if (aimnum==0) {
      free(indx);indx=NULL;
      free(R1);free(R2);R1=R2=NULL;
      free(rad);free(pdel);rad=pdel=NULL;
      return count;
    }
    
    /* set up parameters */
    
    ttoler = 3*aimnum*toler;              /* adjust tolerance */
    key = 1;                              /* initial superstep type */
    m = 1;                                /* Type 1 multiplier */
    sct = 1;                              /* Type 1 count */
    fct = 2;                              /* Type 2 minimum count */
    
    /* do one iteration to get started */  
    c0=h_rif_routine(method_flag);
    if (c0<0.0) goto ABORT;
    if (dbg==1) 
      fprintf(dbugfile,"%4i %1i %8.4e %8.6e 0 \n",count,key,
	      (singledouble)c0,(singledouble)0.0);
    
    if(state != NULL) {
      localstate->aimnum = aimnum;
      localstate->key = key;
      localstate->maxBadCuts = maxBadCuts;
      localstate->minBadCuts = minBadCuts;
      localstate->sumBadCuts = sumBadCuts;
      localstate->cntBadCuts = cntBadCuts;
      localstate->sct = sct;
      localstate->fct = fct;
      localstate->dbugfile = dbugfile;
      localstate->m = m;
      localstate->c0 = c0;
      localstate->ttoler = ttoler;
      localstate->lmax = lmax;
    }
  } /* end of PASS_INIT case */
  
  if(passNo & PASS_MIDDLE) {
    if(state != NULL) {
      localstate = (x_riffle_state *) *state;
      aimnum = localstate->aimnum;
      indx = localstate->index;
      key = localstate->key;
      maxBadCuts = localstate->maxBadCuts;
      minBadCuts = localstate->minBadCuts;
      sumBadCuts = localstate->sumBadCuts;
      cntBadCuts = localstate->cntBadCuts;
      sct = localstate->sct;
      fct = localstate->fct;
      dbugfile = localstate->dbugfile;
      m = localstate->m;
      c0 = localstate->c0;
      ttoler = localstate->ttoler;
      lmax = localstate->lmax;
      R1 = localstate->R1;
      R2 = localstate->R2;
    }
    /* ======== Begin Main Loop ============== */
    while ((c0 >ttoler && count<passes) || (count == 0 && state != NULL)) {
      /* fixup: block write? */
      for (i=1;i<=p->nodecount;i++) R1[i] = rad[i];
      
      numBadCuts = 0;
      do  {   // Make sure factor < 1.0
	c1=h_rif_routine(method_flag);
	if (c1<0.0) goto ABORT;
	factor = c1/c0;
	if (factor >= 1.0) {
	  c0 = c1;
	  key = 1;
	  numBadCuts++;
	}
	if(numBadCuts > MAX_ALLOWABLE_BAD_CUTS) goto ABORT;
      } while(factor >= 1.0);
    
      if (dbg==1) 
	fprintf(dbugfile,"%4i %1i %8.4e %8.6e 0 \n",count,key,
		(singledouble)c1,(singledouble)factor);
      
      if(cntBadCuts == 0)  {
	maxBadCuts = numBadCuts;
	minBadCuts = numBadCuts;
	sumBadCuts = numBadCuts;
      }
      else  {
	maxBadCuts = (numBadCuts > maxBadCuts) ? numBadCuts : maxBadCuts;
	minBadCuts = (numBadCuts < minBadCuts) ? numBadCuts : minBadCuts;
	sumBadCuts += numBadCuts;
      }
      cntBadCuts++;
      
      
      /* superstep calculation */
      
      /* save values */
      for (i=1;i<=p->nodecount;i++) R2[i] = rad[i];
      
      /* compute super step */
      if (key==1) {             /*  type 1  SS */
	lambda = m*factor;
	mmax = 0.75/(1-factor);               /* upper limit on m */
	m = (mmax < (mm=(1+0.8/(sct+1))*m)) ? mmax : mm;
      }
      else {                    /*  type 2 SS */
	if (sct>fct && fabs(factor-fact0)<ftol) {  /* try SS-2 */
	  factor = factor/(1-factor);
	  sct = -1;
	  key = 3;
	}
	lambda = factor;                  /* do something */
      }
      
      if (lambda>lmax/3) {   /* recompute lmax */
	lmax = 10*lambda;
	for (j=0;j<aimnum;j++) {         /* find max step */
	  i = indx[j];
	  r = rad[i];
	  rat = r - R1[i];
	  if (rat>0)
	    lmax = (lmax < (tr=((1-r)/rat))) ? lmax : tr;    
	  /* to keep R<1 */
	  else if (rat < 0)
	    lmax = (lmax < (tr= (-r/rat))) ? lmax : tr;  
	  /* to keep R>0 */
	}
      }
      
      lambda = (lambda>lmax/2) ? lmax/2 : lambda;
    
      /* interpolate new labels */                                            
      for (j=0;j<aimnum;j++) {
	i = indx[j];
	rad[i] += lambda*(rad[i]-R1[i]);
      }
      sct++;
      fact0 = factor;
      
      /* end of superstep */
      
      /* do step/check superstep */
      
      /* check results */
      pred = exp(lambda*log(factor));          /* predicted improvement */
      act = c0/c1;                   /* actual improvement */
      if (act<1) {                   /* did some good */
	if (dbg==1) 
	  fprintf(dbugfile,"%4i %1i %8.4e %8.6e %8.4e \n",
		  count,key,(singledouble)c0,(singledouble)act,
		  (singledouble)lambda);
	if (act>pred) {          /* not as good as expected: reset */
	  m = 1;
	  sct = 0;
	  if (key==1) key = 2;
	}                       /* implied else: accept result */
	if (key==3) key = 1;
      }
      else {                         /* reset to before superstep */
	if (dbg==1) 
	  fprintf(dbugfile,"%4i %1i %8.4e %8.6e %8.4e \n",
		  count,4,(singledouble)c0,(singledouble)act,
		  (singledouble)lambda);
	m = 1;
	sct =0;
	for (i=1;i<=p->nodecount;i++) rad[i]  = R2[i];
	c0 = c1;
	key = 1;
      }
      
      /* show activity */
      if (!(count % 10)) repack_activity_msg("x");
      count++;
    } /* end of main while loop */

    if(state != NULL) {
      localstate->aimnum = aimnum;
      localstate->index = indx;
      localstate->key = key;
      localstate->maxBadCuts = maxBadCuts;
      localstate->minBadCuts = minBadCuts;
      localstate->sumBadCuts = sumBadCuts;
      localstate->cntBadCuts = cntBadCuts;
      localstate->sct = sct;
      localstate->fct = fct;
      localstate->dbugfile = dbugfile;
      localstate->m = m;
      localstate->c0 = c0;
      localstate->ttoler = ttoler;
      localstate->lmax = lmax;
      localstate->R1 = R1;
      localstate->R2 = R2;
    }
  } /* end of PASS_MIDDLE case */
  

  /* fixup: don't understand management of open files in 
     the various situations. Who free's the space? */
  if(passNo & PASS_LAST) {
    if(state != NULL) {
      localstate = (x_riffle_state *) *state;
      aimnum = localstate->aimnum;
      indx = localstate->index;
      key = localstate->key;
      maxBadCuts = localstate->maxBadCuts;
      minBadCuts = localstate->minBadCuts;
      sumBadCuts = localstate->sumBadCuts;
      cntBadCuts = localstate->cntBadCuts;
      sct = localstate->sct;
      fct = localstate->fct;
      dbugfile = localstate->dbugfile;
      m = localstate->m;
      c0 = localstate->c0;
      ttoler = localstate->ttoler;
      lmax = localstate->lmax;
      R1 = localstate->R1;
      R2 = localstate->R2;
      free(localstate);
    }
    if(sumBadCuts > 0)  {
        sprintf(msgbuf, "\nreturning:\n\tmaxBadCuts was "
		"%d\n\tminBadCuts was %d\n"
		"\tavgBadCuts was %.8lf\n\tfinal cut value was %le\n",
		maxBadCuts, minBadCuts, 
		sumBadCuts / (singledouble) cntBadCuts,
		(singledouble) c0);
        msg();
    }    
  } /* end of PASS_LAST case */

  if(cutp != NULL)  *cutp = c0;
  free(indx);indx=NULL;
  free(R1);free(R2);R1=R2=NULL;
  free(pdel);pdel=NULL;
  for (i=1;i<=p->nodecount;i++) h_pR_ptr[i].rad=rad[i];
  free(rad);rad=NULL;
  if (dbg==1 && dbugfile) 
    {fclose(dbugfile);dbugfile=NULL;}
  /* note: angle sums have not been updated */
  return count;

 ABORT:
  if (indx) {free(indx);indx=NULL;}
  if (R1) {free(R1);free(R2);R1=R2=NULL;}
  if (pdel) {free(pdel);pdel=NULL;}
  if (rad) {
    for (i=1;i<=p->nodecount;i++) h_pR_ptr[i].rad=rad[i];
    free(rad);rad=NULL;
  }
  if (dbg==1 && dbugfile) 
    {fclose(dbugfile);dbugfile=NULL;}
  sprintf(msgbuf,"Aborting repack: radius went negative.");
  emsg();
  if(cutp != NULL)  *cutp = c0;
  return 0;
} /* h_riffle_O3 */
 
double h_rif_routine(int method_flag)
  /* Computationally intense part of repack. Return accum error
or -1.0 to abort. 

Method 4 seems to be most efficient. */
{
  int i,j,k,N,*flower;
  double bet,r,rr,r3,r2,sr,m1,m2,m3,onemr,fbest,aae=0;

  for (j=0;j<aimnum;j++) {
    fbest = 0;
    i = indx[j];
    N=h_pK_ptr[i].num;
    if ((r = rad[i])<0) r = (double) 1;
    onemr=1.0-r;
    flower=h_pK_ptr[i].flower;

    switch(method_flag) {
                      
case 1: /* Method 1: original */
  {
    r2 = rad[*flower];
    if (r2<=0) r2=m2=(double) 1;
    else m2=r+onemr*r2;
    for (k=1;k<=N;k++) {  
      r3 = rad[*(flower+k)];
      if (r3<=0) r3=m3=(double) 1;
      else m3=r+onemr*r3;
      fbest += acos(sqrt(r*(r+onemr*(r2+r3-r2*r3))/(m2*m3))); 
      m2 = m3;
      r2 = r3;
    }
    break;
  } 

case 2: /* Method 2: original method if N>10; else, precompute. */
  {
    if (N>10) { 
      r2 = rad[*(flower)];
      if (r2<=0) r2=m2=(double) 1;
      else m2=r+onemr*r2;
      for (k=1;k<=N;k++) { 
	r3 = rad[*(flower+k)];
	if (r3<=0) r3=m3=(double) 1;
	else m3=r+onemr*r3;
	fbest += acos(sqrt(r*(r+onemr*(r2+r3-r2*r3))/(m2*m3))); 
	m2 = m3;
	r2 = r3;
      } 
    }
    else {
      for (k=0;k<=N;k++) {
	locr[k]=rad[*(flower+k)];
	if (locr[k]<=0.0) locr[k]=mval[k]=1.0;
	else  mval[k]=r+onemr*locr[k];
      }
      for (k=1;k<=N;k++) {
	fbest += 
	  acos(sqrt(r*(r+onemr*(locr[k-1]+locr[k]-locr[k-1]*locr[k]))/
		    (mval[k-1]*mval[k])));
      }
    }
    break;
  }

case 3: /* Method 3: cut out loop */
  {
    if (N>10) { 
      r2 = rad[*flower];
      if (r2<=0) r2=m2=(double) 1;
      else m2=r+onemr*r2;
      for (k=1;k<=N;k++) { 
	r3 = rad[*(flower+k)];
	if (r3<=0) r3=m3=(double) 1;
	else m3=r+onemr*r3;
	fbest += acos(sqrt(r*(r+onemr*(r2+r3-r2*r3))/(m2*m3))); 
	m2 = m3;
	r2 = r3;
      } 
    }
    else {
      for (k=0;k<=N;k++) {
	locr[k]=rad[*(flower+k)];
	if (locr[k]<=0.0) locr[k]=mval[k]=1.0;
	else  mval[k]=r+onemr*locr[k];
      }
      switch(N) { /* fall through from different starting points */ 
case 10:
  fbest += acos(sqrt(r*(r+onemr*(locr[9]+locr[10]-locr[9]*locr[10]))/
		    (mval[9]*mval[10])));
case 9:
  fbest += acos(sqrt(r*(r+onemr*(locr[8]+locr[9]-locr[8]*locr[9]))/
		    (mval[8]*mval[9])));
case 8:
  fbest += acos(sqrt(r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/
		    (mval[7]*mval[8])));
case 7:
  fbest += acos(sqrt(r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/
		    (mval[6]*mval[7])));
case 6:
  fbest += acos(sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/
		    (mval[5]*mval[6])));
case 5:
  fbest += acos(sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/
		    (mval[4]*mval[5])));
case 4:
  fbest += acos(sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/
		    (mval[3]*mval[4])));
case 3:
    fbest += acos(sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/
		       (mval[2]*mval[3])));
case 2:
  {
    fbest += acos(sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/
		       (mval[1]*mval[2])));
    fbest += acos(sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/
		       (mval[0]*mval[1])));
    break;
  }
      } /* end of switch on N */
    }
    break;
  }

case 4: /* Method 4: (Seems best.) intermediate result vector */
  {
    if (N>10) { 
      r2 = rad[*flower];
      if (r2<=0) r2=m2=(double) 1;
      else m2=r+onemr*r2;
      for (k=1;k<=N;k++) { 
	r3 = rad[*(flower+k)];
	if (r3<=0) r3=m3=(double) 1;
	else m3=r+onemr*r3;
	fbest += acos(sqrt(r*(r+onemr*(r2+r3-r2*r3))/(m2*m3))); 
	m2 = m3;
	r2 = r3;
      } 
    }
    else {
      for (k=0;k<=N;k++) {
	locr[k]=rad[*(flower+k)];
	if (locr[k]<=0.0) locr[k]=mval[k]=1.0;
	else  mval[k]=r+onemr*locr[k];
      }
      switch(N) { /* fall through from different starting points */ 
case 10:
  {
    rslt[10]=r*(r+onemr*(locr[9]+locr[10]-locr[9]*locr[10]))/(mval[9]*mval[10]);
    rslt[9]=r*(r+onemr*(locr[8]+locr[9]-locr[8]*locr[9]))/(mval[8]*mval[9]);
    rslt[8]=r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]);
    rslt[7]=r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]);
    rslt[6]=r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]);
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 9:
  {
    rslt[9]=r*(r+onemr*(locr[8]+locr[9]-locr[8]*locr[9]))/(mval[8]*mval[9]);
    rslt[8]=r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]);
    rslt[7]=r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]);
    rslt[6]=r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]);
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 8:
  {
    rslt[8]=r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]);
    rslt[7]=r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]);
    rslt[6]=r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]);
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 7:
  {
    rslt[7]=r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]);
    rslt[6]=r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]);
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 6:
  {
    rslt[6]=r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]);
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 5:
  {
    rslt[5]=r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]);
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 4:
  {
    rslt[4]=r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]);
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 3:
  {
    rslt[3]=r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]);
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
case 2:
  {
    rslt[2]=r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]);
    rslt[1]=r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]);
    break;
  }
      } /* end of switch on N */
      
      for (k=1;k<=N;k++) fbest += acos(sqrt(rslt[k]));
    }
    break;
  }
case 5: /* Method 5: intersperse sqrt */
  {
    if (N>10) { 
      r2 = rad[*flower];
      if (r2<=0) r2=m2=(double) 1;
      else m2=r+onemr*r2;
      for (k=1;k<=N;k++) { 
	r3 = rad[*(flower+k)];
	if (r3<=0) r3=m3=(double) 1;
	else m3=r+onemr*r3;
	fbest += acos(sqrt(r*(r+onemr*(r2+r3-r2*r3))/(m2*m3))); 
	m2 = m3;
	r2 = r3;
      } 
    }
    else {
      for (k=0;k<=N;k++) {
	locr[k]=rad[*(flower+k)];
	if (locr[k]<=0.0) locr[k]=mval[k]=1.0;
	else  mval[k]=r+onemr*locr[k];
      }
      switch(N) { /* fall through from different starting points */ 
case 10:
  {
    rslt[10]=
      sqrt(r*(r+onemr*(locr[9]+locr[10]-locr[9]*locr[10]))/(mval[9]*mval[10]));
    rslt[9]=
      sqrt(r*(r+onemr*(locr[8]+locr[9]-locr[8]*locr[9]))/(mval[8]*mval[9]));
    rslt[8]=
      sqrt(r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]));
    rslt[7]=
      sqrt(r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]));
    rslt[6]=
      sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]));
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 9:
  {
    rslt[9]=
      sqrt(r*(r+onemr*(locr[8]+locr[9]-locr[8]*locr[9]))/(mval[8]*mval[9]));
    rslt[8]=
      sqrt(r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]));
    rslt[7]=
      sqrt(r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]));
    rslt[6]=
      sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]));
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 8:
  {
    rslt[8]=
      sqrt(r*(r+onemr*(locr[7]+locr[8]-locr[7]*locr[8]))/(mval[7]*mval[8]));
    rslt[7]=
      sqrt(r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]));
    rslt[6]=
      sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]));
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 7:
  {
    rslt[7]=
      sqrt(r*(r+onemr*(locr[6]+locr[7]-locr[6]*locr[7]))/(mval[6]*mval[7]));
    rslt[6]=
      sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]));
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 6:
  {
    rslt[6]=
      sqrt(r*(r+onemr*(locr[5]+locr[6]-locr[5]*locr[6]))/(mval[5]*mval[6]));
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 5:
  {
    rslt[5]=
      sqrt(r*(r+onemr*(locr[4]+locr[5]-locr[4]*locr[5]))/(mval[4]*mval[5]));
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 4:
  {
    rslt[4]=
      sqrt(r*(r+onemr*(locr[3]+locr[4]-locr[3]*locr[4]))/(mval[3]*mval[4]));
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 3:
  {
    rslt[3]=
      sqrt(r*(r+onemr*(locr[2]+locr[3]-locr[2]*locr[3]))/(mval[2]*mval[3]));
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
case 2:
  {
    rslt[2]=
      sqrt(r*(r+onemr*(locr[1]+locr[2]-locr[1]*locr[2]))/(mval[1]*mval[2]));
    rslt[1]=
      sqrt(r*(r+onemr*(locr[0]+locr[1]-locr[0]*locr[1]))/(mval[0]*mval[1]));
    break;
  }
      } /* end of switch on N */
      
      for (k=1;k<=N;k++) fbest += acos(rslt[k]);
    }
    break;
  }

    } /* end of method switch */

    /* set up for model */
    bet = sin(fbest/N);
    if (r>switchtol) {            /* use hyperbolic UNM */
      r = 1 - r;
      sr = sqrt(r);
      r2 = (bet-sr)/(bet*r-sr); /* reference radius */
      if (r2>0) {                       /* calc new label */
	m1 = 1 - r2;
	m2 = 2*pdel[i];
	m3 = m2/(sqrt(m1*m1+m2*m2*r2)+m1);
	r2 = 1-m3*m3; 
      }
      else
	r2 = 1-pdel[i]*pdel[i];                /* use upper limit */
    }
    else                                 /* use euclidean UNM */
      r2 = r*bet*(1-pdel[i])/(pdel[i]*(1-bet));
    
    rad[i] = r2;          /* store new label */ 
    fbest = 2.0*fbest-h_pR_ptr[i].aim;
    aae += fbest*fbest;           /* accumulate error */
  }
  return (sqrt(aae));
} /* h_rif_routine */
